
from dataclasses import dataclass, field
import matplotlib as mpl
import matplotlib.transforms as mtr
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
import cartopy.crs as ccrs
from matplotlib.gridspec import GridSpec
import matplotlib.axes as maxes
from mpl_toolkits.axes_grid1 import make_axes_locatable
import string
import os

from pathlib import Path, PosixPath
from cartopy.mpl.geoaxes import GeoAxes
from typing import Union, List, Literal

# Hier kann man viel mehr zusammenfassen, da in den
# sub-classen viel wiederholt wird

@dataclass(kw_only = True)
class fig_001:

    fy: float = 6.7
    fx: float = 6.7

    dpi: int = 300
    font_color: str = 'dimgray'
    font_dir: Path = Path('/p/scratch/cjibg31/jibg3105/projects/crusty/src/neoplot/fonts/')
    font: str = 'Montserrat'
    constrained: bool = True
    axs: list[GeoAxes] = field(default_factory = list)

    def create(self):
        
        font_files = fm.findSystemFonts(fontpaths = [self.font_dir])
    
        for font_file in font_files:
            fm.fontManager.addfont(font_file)

        mpl.rcParams['font.family'] = self.font
        mpl.rcParams['xtick.color'] = self.font_color
        mpl.rcParams['ytick.color'] = self.font_color

        fig_ = plt.figure(figsize = (self.fx, self.fy), 
                          constrained_layout = self.constrained,
                          dpi = self.dpi)
        
        self.fig = fig_
        
        return self
    
    def annotation(self, 
                   x : float = 0.05, 
                   y : float = 1.05, 
                   fs: float = 12.0):

        abc = string.ascii_lowercase
    
        list_abc = list(abc)

        for iax, ax in enumerate(self.axs):

            ax.text(x, y, list_abc[iax] + ')', 
                    fontsize = fs,
                    transform = ax.transAxes, 
                    va = 'bottom', 
                    ha = 'center')

    def save(self, 
             path: os.PathLike,
             fformat: str = 'png'):

        self.fig.savefig(f'{path}.{fformat}',
                         dpi = self.dpi,
                         bbox_inches = 'tight')
        
        plt.close()

        

@dataclass(kw_only = True)
class single_001(fig_001):

    color_bar: None | list[str] = None
    projection: mtr.Transform | ccrs.Projection | \
                None | Literal['EU3'] = None
    frame: bool = False


    def create(self):

        super().create()

        nrows = 1
        ncols = 1
        axs = []

        if self.projection == 'EU3':
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                           semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                   pole_latitude = rotnpole_lat,
                                   globe = globe)
            
            self.projection = self.rp


        gs = GridSpec(figure = self.fig, 
                      ncols = ncols, 
                      nrows = nrows)

        axs.append(self.fig.add_subplot(gs[0, 0], 
                                        projection = self.projection, 
                                        frameon = self.frame))

        
        self.axs = axs

        return self

    


@dataclass(kw_only = True)
class triple_001(fig_001):
    
    projection: List[Union[mtr.Transform, ccrs.Projection, Literal['EU3'], None]] | \
                    mtr.Transform | ccrs.Projection | Literal['EU3'] | None = None
    frame: bool = False

    def create(self):

        super().create()

        axs = []

        nrows = 1
        ncols = 3

        if not isinstance(self.projection, list): 
            
            self.projection = [self.projection for i in range(ncols)]

        if 'EU3' in self.projection:
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                               semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                       pole_latitude = rotnpole_lat,
                                       globe = globe)

        
            self.active_proj = [self.rp if p == 'EU3' else p 
                                for p in self.projection]
            
        else:
            self.active_proj = self.projection

        gs = GridSpec(figure = self.fig, 
                      nrows = nrows,
                      ncols = ncols,)

        for iax in range(ncols):

            axs.append(self.fig.add_subplot(gs[0, iax], 
                                            projection = self.active_proj[iax], 
                                            frameon = self.frame))

        self.axs = axs

        return self


@dataclass(kw_only = True)      
class double_001(fig_001):
    
    projection: List[Union[mtr.Transform, ccrs.Projection, Literal['EU3'], None]] | \
                    mtr.Transform | ccrs.Projection | Literal['EU3'] | None = None
    frame: bool = False


    def create(self):

        super().create()

        axs = []

        nrows = 1
        ncols = 2

        if not isinstance(self.projection, list): 
            
            self.projection = [self.projection for i in range(ncols)]

        if 'EU3' in self.projection:
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                               semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                       pole_latitude = rotnpole_lat,
                                       globe = globe)

        
            self.active_proj = [self.rp if p == 'EU3' else p 
                                for p in self.projection]
            
        else:
            self.active_proj = self.projection

        gs = GridSpec(figure = self.fig, 
                      nrows = nrows,
                      ncols = ncols)

        for iax in range(ncols):

            axs.append(self.fig.add_subplot(gs[0, iax], 
                                            projection = self.active_proj[iax], 
                                            frameon = self.frame))

        self.axs = axs

        return self
       
            
@dataclass(kw_only = True)
class quatro_001(fig_001):
    
    projection: List[Union[mtr.Transform, ccrs.Projection, Literal['EU3'], None]] | \
                    mtr.Transform | ccrs.Projection | Literal['EU3'] | None = None
    frame: bool = False

    wspace: float = 0.1
    hspace: float = 0.05

    def create(self):

        super().create()

        axs = []

        nrows = 2
        ncols = 2

        if not isinstance(self.projection, list): 
            
            self.projection = [self.projection for i in range(ncols*nrows)]

        if 'EU3' in self.projection:
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                               semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                       pole_latitude = rotnpole_lat,
                                       globe = globe)

        
            self.active_proj = [self.rp if p == 'EU3' else p 
                                for p in self.projection]
            
        else:
            self.active_proj = self.projection

        gs = GridSpec(figure = self.fig, 
                      nrows = nrows,
                      ncols = ncols,
                      wspace = self.wspace,
                      hspace = self.hspace)
        
        positions = [gs[0, 0], gs[0, 1], gs[1, 0], gs[1, 1]]

        for iax, ax in enumerate(positions):

            axs.append(self.fig.add_subplot(ax, 
                                            projection = self.active_proj[iax], 
                                            frameon = self.frame))

        self.axs = axs

        return self
        

@dataclass(kw_only = True)
class twoxfour(fig_001):

    projection: List[Union[mtr.Transform, ccrs.Projection, Literal['EU3'], None]] | \
                    mtr.Transform | ccrs.Projection | Literal['EU3'] | None = None
    frame: bool = False

    wspace: float = 0.1
    hspace: float = 0.05

    def create(self):

        super().create()

        nrows = 2
        ncols = 4

        if not isinstance(self.projection, list): 
            
            self.projection = [self.projection for i in range(ncols*nrows)]

        if 'EU3' in self.projection:
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                               semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                       pole_latitude = rotnpole_lat,
                                       globe = globe)

        
            self.active_proj = [self.rp if p == 'EU3' else p 
                                for p in self.projection]
            
        else:
            self.active_proj = self.projection

        gs = GridSpec(figure = self.fig, 
                      nrows = nrows,
                      ncols = ncols,
                      wspace = self.wspace,
                      hspace = self.hspace)
        
        positions = [gs[0, 0], gs[0, 1], gs[0, 2], gs[0, 3],
                     gs[1, 0], gs[1, 1], gs[1, 2], gs[1, 3],]
        
        self.positions = positions

        for iax, ax in enumerate(positions):

            self.axs.append(self.fig.add_subplot(ax, 
                                        projection = self.active_proj[iax], 
                                        frameon = self.frame))

        return self


@dataclass(kw_only = True)
class fourxtwo(fig_001):

    projection: List[Union[mtr.Transform, ccrs.Projection, Literal['EU3'], None]] | \
                    mtr.Transform | ccrs.Projection | Literal['EU3'] | None = None
    frame: bool = False

    wspace: float = 0.1
    hspace: float = 0.05

    def create(self):

        super().create()

        nrows = 4
        ncols = 2

        if not isinstance(self.projection, list): 
            
            self.projection = [self.projection for i in range(ncols*nrows)]

        if 'EU3' in self.projection:
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                               semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                       pole_latitude = rotnpole_lat,
                                       globe = globe)

        
            self.active_proj = [self.rp if p == 'EU3' else p 
                                for p in self.projection]
            
        else:
            self.active_proj = self.projection

        gs = GridSpec(figure = self.fig, 
                      nrows = nrows,
                      ncols = ncols,
                      wspace = self.wspace,
                      hspace = self.hspace)
        
        positions = [gs[0, 0], gs[0, 1], 
                     gs[1, 0], gs[1, 1],
                     gs[2, 0], gs[2, 1], 
                     gs[3, 0], gs[3, 1],]
        
        self.positions = positions

        for iax, ax in enumerate(positions):

            self.axs.append(self.fig.add_subplot(ax, 
                                        projection = self.active_proj[iax], 
                                        frameon = self.frame))

        return self

@dataclass(kw_only = True)
class fivexthree(fig_001):
    
    projection: List[Union[mtr.Transform, ccrs.Projection, Literal['EU3'], None]] | \
                    mtr.Transform | ccrs.Projection | Literal['EU3'] | None = None
    frame: bool = False

    wspace: float = 0.1
    hspace: float = 0.05

    def create(self):

        super().create()

        nrows = 5
        ncols = 3

        if not isinstance(self.projection, list): 
            
            self.projection = [self.projection for i in range(ncols*nrows)]

        if 'EU3' in self.projection:
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                               semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                       pole_latitude = rotnpole_lat,
                                       globe = globe)

        
            self.active_proj = [self.rp if p == 'EU3' else p 
                                for p in self.projection]
            
        else:
            self.active_proj = self.projection

        gs = GridSpec(figure = self.fig, 
                      nrows = nrows,
                      ncols = ncols,
                      wspace = self.wspace,
                      hspace = self.hspace)
        
        positions = [gs[0, 0], gs[0, 1], gs[0, 2],
                     gs[1, 0], gs[1, 1], gs[1, 2],
                     gs[2, 0], gs[2, 1], gs[2, 2],
                     gs[3, 0], gs[3, 1], gs[3, 2],
                     gs[4, 0], gs[4, 1], gs[4, 2]]
        
        self.positions = positions

        for iax, ax in enumerate(positions):

            self.axs.append(self.fig.add_subplot(ax, 
                                        projection = self.active_proj[iax], 
                                        frameon = self.frame))

        return self
    


@dataclass(kw_only = True)
class fivexfour(fig_001):
    
    projection: List[Union[mtr.Transform, ccrs.Projection, Literal['EU3'], None]] | \
                    mtr.Transform | ccrs.Projection | Literal['EU3'] | None = None
    frame: bool = False

    wspace: float = 0.1
    hspace: float = 0.05

    def create(self):

        super().create()

        nrows = 5
        ncols = 4

        if not isinstance(self.projection, list): 
            
            self.projection = [self.projection for i in range(ncols*nrows)]

        if 'EU3' in self.projection:
    
            rotnpole_lat: float = 39.25 
            rotnpole_lon: float = -162.0

            semmj_axis: None | int = 6370000
            semmn_axis: None | int = 6370000 

            globe = ccrs.Globe(semimajor_axis = semmj_axis,
                               semiminor_axis = semmn_axis)

            self.rp = ccrs.RotatedPole(pole_longitude = rotnpole_lon,
                                       pole_latitude = rotnpole_lat,
                                       globe = globe)

        
            self.active_proj = [self.rp if p == 'EU3' else p 
                                for p in self.projection]
            
        else:
            self.active_proj = self.projection

        gs = GridSpec(figure = self.fig, 
                      nrows = nrows,
                      ncols = ncols,
                      wspace = self.wspace,
                      hspace = self.hspace)
        
        positions = [gs[0, 0], gs[0, 1], gs[0, 2], gs[0, 3],
                     gs[1, 0], gs[1, 1], gs[1, 2], gs[1, 3],
                     gs[2, 0], gs[2, 1], gs[2, 2], gs[2, 3],
                     gs[3, 0], gs[3, 1], gs[3, 2], gs[3, 3],
                     gs[4, 0], gs[4, 1], gs[4, 2], gs[4, 3]]
        
        self.positions = positions

        for iax, ax in enumerate(positions):

            self.axs.append(self.fig.add_subplot(ax, 
                                        projection = self.active_proj[iax], 
                                        frameon = self.frame))

        return self



def keys_same_value(idict: dict):

    from itertools import groupby

    grouped                 = groupby(sorted(idict, key = idict.get), key = idict.get)

    dict_grouped            = {v: list(ks) for v, ks in grouped}

    return dict_grouped


def translate_dict_values(idict: dict, idict_translate: dict):

    idict_translated        = {}

    for k, v in idict.items():
        
        if isinstance(v, list):
            
            idict_translated.update({k: [idict_translate[vv] for vv in v]})

        else:

            idict_translated.update({k: idict_translate[v]})

    return idict_translated


def filter_dict_values(idict, ilist):

    idict_filtered          = {}

    for k, v in idict.items():
        
        if isinstance(v, list):
            
            idict_filtered.update({k: [vv for vv in v if vv in ilist]})

        else:

            if v in ilist:

                idict_filtered.update({k: v})

    return idict_filtered


def filter_dict_keys(idict, ilist):

    idict_filtered          = {k: v for k, v in idict.items() if k in ilist}

    return idict_filtered


def list_to_array(ilist):

    import numpy as np

    array_T                 = np.array(ilist)

    array                   = array_T.T

    return array


import matplotlib as mpl
import matplotlib.pyplot as plt
import cartopy.crs as ccrs


def horizontal_right_map_two_cax(fx: float = 6.7,
                                 fy: float = 6.7,
                                 dpi: int = 300, 
                                 layout: str = 'constrained', 
                                 projection: None | ccrs.Projection = None, 
                                 frame: bool = False,
                                 hspace: float = 0.05, 
                                 wspace: float = 0.05,
                                 annotation: bool = True, 
                                 x_an: float = -0.05, 
                                 y_an: float = 1.05): 

    from matplotlib.gridspec import GridSpec
    from matplotlib.patches import ConnectionPatch

    from my_.plot.style import style_1
    from my_.plot.init_ax import init_annotation_ax

    style_1()

    ncols, nrows                    = 2, 6

    figure                          = plt.figure(figsize = (fx, fy), 
                                                 dpi = dpi, 
                                                 layout = layout)

    gs                              = GridSpec(figure = figure, 
                                               ncols = ncols, 
                                               nrows = nrows,
                                               height_ratios = [1, 15, 4, 1.5, 11, 11],
                                               width_ratios = [10, 10],
                                               hspace = hspace, 
                                               wspace = wspace)
    

    
    axp_kw                          = {'frameon': frame}
    axm_kw                          = {'frameon': frame, 'projection': projection}

    axp1                            = plt.subplot(gs[1, 0], **axp_kw)
    axp2                            = plt.subplot(gs[1, 1], **axp_kw)
    axm                             = plt.subplot(gs[4:6, 0:2], **axm_kw)
    
    axl1                            = plt.subplot(gs[0, :2], **axp_kw)
    axl2                            = plt.subplot(gs[3, 0], **axp_kw)
    axc                             = plt.subplot(gs[3, 1], **axp_kw)

    transFigure                     = figure.transFigure.inverted()

    line                            = ConnectionPatch(xyA=(0, 0), xyB=(0, 1), coordsA="axes fraction", coordsB="axes fraction",
                                                    axesA=axp1, axesB=axp2, color="red")
    
    axp2.add_artist(line)

    axs                             = [axp1, axp2, axm]

    if annotation: init_annotation_ax(axs, x = x_an, y = y_an)

    return figure, axs, axl1, axl2, axc


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import colorcet as cc

def map_EU3_point_locations_lc_hclim(lats, lons, landcover, hydroclimate, rotnpole_lat: float, rotnpole_lon: float, semmj_axis: int, semmn_axis: int,
                        lon_extents: list, lat_extents: list, lw_grid: float = 0.5, lw_coast: float = 0.5,  color_grid: str = 'gray',
                        ls_grid: str = '--', xticks: list = [], yticks: list = [], fs_label: float = 10, size_marker: float = 4.0, 
                        marker: str = 'x', color_marker: str = 'firebrick', alpha: float = 0.85, zorder: int = 5, title: str = '',
                        marker_legend_args: dict = {}):

    from my_.figures.single import square_top_right_cax
    from my_.plot.style import style_1
    from my_.plot.maps import EU3_plot_lines, map_point_locations
    from my_.plot.init_ax import EU3_plot_init
    from my_.plot.colors import colorbar
    from my_.plot.legend import marker_legend
    from user_in.options_analyses import selected_landcover, markers
    
    import matplotlib.pyplot as plt
    import matplotlib as mpl

    print('Plot locations map...\n')

    style_1()
    
    rp, pc, xs, ys              = EU3_plot_init(rotnpole_lat, rotnpole_lon, semmj_axis, semmn_axis, lon_extents, lat_extents)

    figure, ax, caxt, caxr      = square_top_right_cax(projection = rp, fy = 5.0)

    ax.set_title(title)
    
    sizes                       = [size_marker] * len(lats)

    cmap                        = plt.get_cmap('BrBG', 6)

    EU3_plot_lines(ax, pc, xs, ys, lw_grid, lw_coast, color_grid, ls_grid, xticks, yticks, fs_label)

    for lc in landcover.unique():
        
        if lc not in selected_landcover: continue

        lat_lc                  = lats.where(landcover == lc)
        lon_lc                  = lons.where(landcover == lc)
        hclim_lc                = hydroclimate.where(landcover == lc)

        colors                  = [cmap(hc/6) for hc in hclim_lc]

        map_point_locations(ax, 
                            lat_lc, 
                            lon_lc, 
                            sizes, 
                            marker = markers[lc], 
                            color = 'None',
                            edgecolor = colors, 
                            projection = pc, 
                            alpha = alpha, 
                            zorder = zorder)

        a = mpl.cm.ScalarMappable(norm = mpl.colors.Normalize(vmin=0, vmax=5), cmap = cmap) 
        ticks = ['Very arid', 'Arid', 'Semi arid', 'Semi humid', 'Humid', 'Very humid']

    colorbar(caxr, a, '', pad = 10, extend = 'neither', fs_label = 12, tick_labels = ticks)
    marker_legend(caxt, markers, **marker_legend_args)

    return figure


def map_EU3_point_locations_hclim(lats, lons, hydroclimate, rotnpole_lat: float, rotnpole_lon: float, semmj_axis: int, semmn_axis: int,
                        lon_extents: list, lat_extents: list, lw_grid: float = 0.5, lw_coast: float = 0.5,  color_grid: str = 'gray',
                        ls_grid: str = '--', xticks: list = [], yticks: list = [], fs_label: float = 10, size_marker: float = 45, 
                        marker: str = 'o', alpha: float = 0.85, zorder: int = 5, title: str = ''):

    from my_.figures.single import square_right_cax
    from my_.plot.style import style_1
    from my_.plot.maps import EU3_plot_lines, map_point_locations
    from my_.plot.init_ax import EU3_plot_init
    from my_.plot.colors import colorbar
    
    import matplotlib.pyplot as plt
    import matplotlib as mpl

    print('Plot locations map...\n')

    style_1()
    
    rp, pc, xs, ys              = EU3_plot_init(rotnpole_lat, rotnpole_lon, semmj_axis, semmn_axis, lon_extents, lat_extents)

    figure, ax, cax             = square_right_cax(projection = rp, fy = 5.0)

    ax.set_title(title)
    
    sizes                       = [size_marker] * len(lats)

    cmap                        = plt.get_cmap('BrBG', 6)

    EU3_plot_lines(ax, pc, xs, ys, lw_grid, lw_coast, color_grid, ls_grid, xticks, yticks, fs_label)
    
    colors                  = [cmap(hc/6) for hc in hydroclimate]

    map_point_locations(ax, lats, lons, sizes, marker = marker, color = colors, projection = pc, alpha = alpha, zorder = zorder)

    a = mpl.cm.ScalarMappable(norm = mpl.colors.Normalize(vmin=0, vmax=5), cmap = cmap) 
    ticks = ['Very arid', 'Arid', 'Semi arid', 'Semi humid', 'Humid', 'Very humid']

    colorbar(cax, a, '', pad = 10, extend = 'neither', fs_label = 12, tick_labels = ticks)

    return figure


def map_EU3_point_locations(lats, lons, rotnpole_lat: float, rotnpole_lon: float, semmj_axis: int, semmn_axis: int,
                        lon_extents: list, lat_extents: list, lw_grid: float = 0.5, lw_coast: float = 0.5,  color_grid: str = 'gray',
                        ls_grid: str = '--', xticks: list = [], yticks: list = [], fs_label: float = 10, size_marker: float = 4.0, 
                        marker: str = 'x', color_marker: str = 'firebrick', alpha: float = 0.7, zorder: int = 5, title: str = ''):

    from my_.figures.single import square
    from my_.plot.style import style_1
    from my_.plot.maps import EU3_plot_lines, map_point_locations
    from my_.plot.init_ax import EU3_plot_init

    print('Plot locations map...\n')

    style_1()
    
    rp, pc, xs, ys              = EU3_plot_init(rotnpole_lat, rotnpole_lon, semmj_axis, semmn_axis, lon_extents, lat_extents)

    figure, ax                  = square(projection = rp)

    ax.set_title(title)
    
    sizes                       = [size_marker] * len(lats)

    EU3_plot_lines(ax, pc, xs, ys, lw_grid, lw_coast, color_grid, ls_grid, xticks, yticks, fs_label)

    map_point_locations(ax, 
                        lats, 
                        lons, 
                        sizes, 
                        marker = marker, 
                        color = 'w',
                        edgecolor = color_marker, 
                        projection = pc, 
                        alpha = alpha, 
                        zorder = zorder)

    

    return figure


def double_EU3_mesh_div_cbar(arrays, lat, lon,
                            rotnpole_lat: float, rotnpole_lon: float, semmj_axis: int, semmn_axis: int,
                            lon_extents: list, lat_extents: list,
                            lw_grid: float = 0.5, lw_coast: float = 0.5,  color_grid: str = 'gray',
                            ls_grid: str = '--', xticks: list = [], yticks: list = [], fs_label: float = 10,
                            cmap: str = 'coolwarm_r', cmap_n: int = 1000, v0: float = None, v1: float = None, 
                            extend: str = 'both', title: str = '', subtitles: list = [], clabel: str = '',
                            fs_title: float = 12, fs_subtitle: float = 10, fs_cbar_label: float = 10):
    

    from my_.plot.style import style_1
    from my_.figures.double import horizontal_cax
    from my_.plot.basic import colormesh
    from my_.plot.maps import EU3_plot_lines
    from my_.plot.init_ax import EU3_plot_init
    from my_.plot.colors import colormap, colorbar

    style_1()

    rp, pc, xs, ys              = EU3_plot_init(rotnpole_lat, rotnpole_lon, semmj_axis, semmn_axis, lon_extents, lat_extents)

    fig, axes, cax              = horizontal_cax(projection = rp)

    label_sides                 = [['bottom'], ['bottom', 'right']]

    cmap_c                      = colormap(cmap, cmap_n)

    fig.suptitle(title, fontsize = fs_title)

    for i_ax, ax in enumerate(axes):

        ax.set_title(subtitles[i_ax], fontsize = fs_subtitle)
        
        EU3_plot_lines(ax, pc, xs, ys, lw_grid, lw_coast, color_grid, ls_grid, xticks, yticks,
                       fs_label, label_sides = label_sides[i_ax])

        artist                  = colormesh(ax, lon, lat, arrays[i_ax], cmap_c, v0, v1, projection = pc)

    colorbar(cax, artist, clabel, pad = 10, extend = extend, fs_label = fs_cbar_label)

    return fig

def single_EU3_mesh_div_cbar(array, lat, lon,
                            rotnpole_lat: float, rotnpole_lon: float, semmj_axis: int, semmn_axis: int,
                            lon_extents: list, lat_extents: list,
                            lw_grid: float = 0.5, lw_coast: float = 0.5,  color_grid: str = 'gray',
                            ls_grid: str = '--', xticks: list = [], yticks: list = [], fs_label: float = 10,
                            cmap: str = 'coolwarm_r', cmap_n: int = 1000, v0: float = None, v1: float = None, 
                            extend: str = 'both', title: str = '', subtitle: str = '', clabel: str = '',
                            fs_title: float = 12, fs_subtitle: float = 10, fs_cbar_label: float = 10):
    

    from my_.plot.style import style_1
    from my_.figures.single import square_right_cax
    from my_.plot.basic import colormesh
    from my_.plot.maps import EU3_plot_lines
    from my_.plot.init_ax import EU3_plot_init
    from my_.plot.colors import colormap, colorbar

    rp, pc, xs, ys              = EU3_plot_init(rotnpole_lat, rotnpole_lon, semmj_axis, semmn_axis, lon_extents, lat_extents)

    fig, ax, cax                = square_right_cax(projection = rp)

    label_sides                 = ['bottom', 'right']

    cmap_c                      = colormap(cmap, cmap_n)

    fig.suptitle(title, fontsize = fs_title)


    ax.set_title(subtitle, fontsize = fs_subtitle)
        
    EU3_plot_lines(ax, pc, xs, ys, lw_grid, lw_coast, color_grid, ls_grid, xticks, yticks,
                    fs_label, label_sides = label_sides)

    artist                  = colormesh(ax, lon, lat, array, cmap_c, v0, v1, projection = pc)

    colorbar(cax, artist, clabel, pad = 10, extend = extend, fs_label = fs_cbar_label)

    return fig

def pie_location_lc_clim_map(df_static: pd.DataFrame,
                            lats: np.ndarray | list | tuple | pd.Series,
                            lons: np.ndarray | list | tuple | pd.Series,
                            landcover: np.ndarray | list | tuple | pd.Series,
                            hydroclimate: np.ndarray | list | tuple | pd.Series,
                            basemap_args: dict, 
                            baselines_args: dict, 
                            map_args: dict,
                            markers: dict,
                            marker_legend_args: dict,
                            selected_landcover: list[str], 
                            colors: dict,
                            color_legend_args:dict) -> plt.figure:
    
    from my_.figures.one_by_three import horizontal_right_map_two_cax
    from my_.series.aggregate import concat
    from my_.plot.basic import pie
    from my_.plot.init_ax import init_pie, EU3_plot_init
    from my_.plot.legend import color_legend, marker_legend
    from my_.plot.colors import colorbar
    from my_.resources.sources import query_static
    from my_.plot.maps import EU3_plot_lines, map_point_locations

    rp, pc, xs, ys              = EU3_plot_init(**basemap_args)

    fig, axs, axl1, axl2, axc   = horizontal_right_map_two_cax(projection = rp)

    landcover                   = df_static['landcover'].where(df_static['landcover'].isin(selected_landcover))

    pct_crop                    = df_static['PCT_CROP'] / 100

    pct_nat                     = df_static['PCT_NATVEG'] / 100

    pft_cols                    = [col for col in df_static.columns if 'PCT_NAT_PFT_' in col]

    cft_cols                    = [col for col in df_static.columns if 'PCT_CFT_' in col]

    pft_rel                     = df_static[pft_cols].multiply(pct_nat, axis = 0)

    cft_rel                     = df_static[cft_cols].multiply(pct_crop, axis = 0)

    df_tot                      = concat([pft_rel, cft_rel], sort = False)

    ind_landcover               = query_static('CLM5-EU3-surf', 'sel_agg_layer_PFT')

    shares_surf                 = [df_tot.iloc[:, ind_landcover[lc]].sum(axis = 1).mean() for lc in selected_landcover]

    shares_surf_o               = shares_surf + [100 - sum(shares_surf)] 

    shares_lc                   = landcover.value_counts(normalize = True)

    cmapc                       = plt.get_cmap('Set2').colors

    colors                      = {k: v for k,v in colors.items()}
    colorsc                     = [cmapc[colors[lc]] for lc in list(selected_landcover)]

    colors['other']             = 7

    colorsc_o                   = colorsc + [cmapc[7]]

    color_legend(axl1, colors, cmapc, **color_legend_args)

    init_pie(axs[0], ax_tag = 'ICOS network')
    init_pie(axs[1], ax_tag = 'CLM5 surface')

    pie(axs[0], shares_lc, colorsc, autopct='%1.1f%%')
    pie(axs[1], shares_surf_o, colorsc_o, autopct='%1.1f%%')

    sizes                       = [map_args['size_marker']] * len(lats)

    cmap                        = plt.get_cmap('BrBG', 6)

    EU3_plot_lines(axs[-1], pc, xs, ys, **baselines_args)
    
    for lc in landcover.unique():
        
        if lc not in selected_landcover: continue

        lat_lc                  = lats.where(landcover == lc)
        lon_lc                  = lons.where(landcover == lc)
        hclim_lc                = hydroclimate.where(landcover == lc)

        colors                  = [cmap(hc/6) for hc in hclim_lc]

        map_point_locations(axs[2], 
                            lat_lc, 
                            lon_lc, 
                            sizes, 
                            marker = markers[lc], 
                            color = 'w', #colors, 
                            edgecolor = colors,
                            projection = pc, 
                            alpha = map_args['alpha'], 
                            zorder = map_args['zorder'])

        a = mpl.cm.ScalarMappable(norm = mpl.colors.Normalize(vmin=0, vmax=5), cmap = cmap) 
        ticks = ['Very\narid', 'Arid', 'Semi\narid', 'Semi\nhumid', 'Humid', 'Very\nhumid']

    colorbar(axc, a, '', pad = 10, extend = 'neither', fs_label = 8, tick_labels = ticks, orientation = 'horizontal')
    marker_legend(axl2, markers, **marker_legend_args)

    return fig



def single_EU3_mesh_cat_cbar(array, lat, lon,
                            rotnpole_lat: float, rotnpole_lon: float, semmj_axis: int, semmn_axis: int,
                            lon_extents: list, lat_extents: list,
                            lw_grid: float = 0.5, lw_coast: float = 0.5,  color_grid: str = 'gray',
                            ls_grid: str = '--', xticks: list = [], yticks: list = [], fs_label: float = 10,
                            cmap: str = 'cet_glasbey_hv', extend: str = 'neither', title: str = '', clabel: str = '', 
                            fs_title: float = 12, fs_cbar_label: float = 10, cbar_tick_labels: list = []): 

    import numpy as np
    from my_.plot.style import style_1
    from my_.plot.maps import EU3_plot_lines
    from my_.plot.init_ax import EU3_plot_init
    from my_.figures.single import square_right_cax
    from my_.plot.colors import colormap, colorbar
    from my_.plot.basic import colormesh

    import cartopy.feature as cfeature

    label_sides                 =  ['bottom', 'right']

    style_1()

    rp, pc, xs, ys              = EU3_plot_init(rotnpole_lat, rotnpole_lon, semmj_axis, semmn_axis, lon_extents, lat_extents)

    figure, ax, cax             = square_right_cax(projection = rp)

    ax.add_feature(cfeature.OCEAN, edgecolor='face', facecolor='white', zorder = 2)

    cmap_n                      = np.max(array) + 1 # "maybe better": length of unique values. But not sure

    cmap_c                      = colormap(cmap, cmap_n)

    ax.set_title(title, fontsize = fs_title)

    EU3_plot_lines(ax, pc, xs, ys, lw_grid, lw_coast, color_grid, ls_grid, xticks, yticks,
                       fs_label, label_sides = label_sides)

    artist                      = colormesh(ax, lon, lat, array, cmap_c, projection = pc)

    colorbar(cax, artist, clabel, pad = 10, extend = extend, 
            fs_label = fs_cbar_label, tick_labels = cbar_tick_labels)

    return figure


def single_EU3_mesh_cont_cbar(array, lat, lon,
                            rotnpole_lat: float, rotnpole_lon: float, semmj_axis: int, semmn_axis: int,
                            lon_extents: list, lat_extents: list,
                            lw_grid: float = 0.5, lw_coast: float = 0.5,  color_grid: str = 'gray',
                            ls_grid: str = '--', xticks: list = [], yticks: list = [], fs_label: float = 10,
                            cmap: str = 'Greens', extend: str = 'neither', title: str = '', clabel: str = '', 
                            fs_title: float = 12, fs_cbar_label: float = 10, cbar_tick_labels: list = []): 

    import numpy as np
    from my_.plot.style import style_1
    from my_.plot.maps import EU3_plot_lines
    from my_.plot.init_ax import EU3_plot_init
    from my_.figures.single import square_right_cax
    from my_.plot.colors import colormap, colorbar
    from my_.plot.basic import colormesh

    import cartopy.feature as cfeature

    label_sides                 =  ['bottom', 'right']

    style_1()

    rp, pc, xs, ys              = EU3_plot_init(rotnpole_lat, rotnpole_lon, semmj_axis, semmn_axis, lon_extents, lat_extents)

    figure, ax, cax             = square_right_cax(projection = rp)

    ax.add_feature(cfeature.OCEAN, edgecolor='face', facecolor='white', zorder = 2)

    cmap_n                      = np.max(array) + 1 # "maybe better": length of unique values. But not sure

    cmap_c                      = colormap(cmap, cmap_n)

    ax.set_title(title, fontsize = fs_title)

    EU3_plot_lines(ax, pc, xs, ys, lw_grid, lw_coast, color_grid, ls_grid, xticks, yticks,
                       fs_label, label_sides = label_sides)

    artist                      = colormesh(ax, lon, lat, array, cmap_c, projection = pc)

    colorbar(cax, artist, clabel, pad = 10, extend = extend, 
            fs_label = fs_cbar_label, tick_labels = cbar_tick_labels)

    return figure


def taylor_landcover(df: pd.DataFrame,
                     obs: str,
                     variables: list[str],
                     colors: dict,
                     markers: dict,
                     landcover: list[str],
                     std_range: tuple = (0, 1.5),
                     marker_legend_args: dict = {},
                     color_legend_args: dict = {}):
    
    from my_.series.group import select_multi_index, nona_level
    from my_.plot.legend import marker_legend, color_legend
    from my_.series.aggregate import single_level_wise, groupwise
    from my_.math.stats import rmse, r, pbias
    from matplotlib.projections import PolarAxes
    from mpl_toolkits.axisartist import floating_axes, grid_finder
    from my_.plot.style import style_1
    from matplotlib.gridspec import GridSpec
    from my_.plot.limits import axgrid
    from my_.resources.sources import query_variables
    from sklearn.preprocessing import MinMaxScaler

    scaler = MinMaxScaler(feature_range = (0, 1))

    print('Plot Taylor diagram...\n')

    cmapc = cc.glasbey_hv[:] + [[0.0, 0.0, 0.0]]

    sources = df.columns.unique(level = 'Source')

    labels_sources = {s: query_variables(s, 'name_label') 
                      for s in sources}

    colorscc = {labels_sources[k]: v 
              for k, v in colors.items() if k in sources}

    stations_uq = df.columns.unique(level = 'Station')
    
    stations_count = df.columns\
                       .get_level_values('Station')\
                       .value_counts()

    
    for s in stations_uq:
        
        if stations_count[s] == 1:

            df = df.drop(s, axis = 1, level = 'Station')
        
    df_nona = nona_level(df, 
                         ['Variable', 'Station'],
                         axis = 1,
                         how = 'any')

    df_n = select_multi_index(df_nona, 
                              ['Variable', 'Landcover'],
                              keys = [variables, landcover],
                              axis = 1)
    
    for var in variables: 

        var_units = query_variables(sources[0], 
                                'var_units')[var]

        var_df = select_multi_index(df_n,
                                    levels = ['Variable'],
                                    keys = [var],
                                    axis = 1)
        
        std_df = var_df.T.groupby(level = ['Landcover', 
                                            'Source'])\
                         .apply(groupwise,
                                [pd.DataFrame.to_numpy,
                                 np.nanstd])

        r_df = var_df.T.groupby(level = ['Landcover'])\
                       .apply(single_level_wise,
                              level = 'Source',
                              key = obs,
                              ffunc = r,
                              axis = 0)

        pbias_df = var_df.T.groupby(level = ['Landcover'])\
                       .apply(single_level_wise,
                              level = 'Source',
                              key = obs,
                              ffunc = pbias,
                              axis = 0)
        
        rmse_df = var_df.T.groupby(level = ['Landcover'])\
                          .apply(single_level_wise,
                                 level = 'Source',
                                 key = obs,
                                 ffunc = rmse,
                                 axis = 0)

        ref_std = select_multi_index(std_df,
                                     levels = ['Source'],
                                     keys = [obs],
                                     axis  = 0)

        tr = PolarAxes.PolarTransform()

        pbias_ms_scale = scaler.fit_transform(np.abs(pbias_df))

        pbias_ms = pbias_df.copy()

        pbias_ms.iloc[:, :] = 6 + np.abs(pbias_ms_scale) * 10
    
        if (r_df < 0).values.any():
            
            extend = True
            tmax = np.pi
            rlocs = [-1, -0.99, -0.95, -0.9, -0.8, -0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6, 0.8, 0.9 ,0.95, 0.99, 1]

        else:
            
            extend = False
            tmax = np.pi/2
            rlocs = [0, 0.2, 0.4, 0.6, 0.8, 0.9 ,0.95, 0.99, 1]

        tlocs = np.arccos(rlocs)

        gl1 = grid_finder.FixedLocator(tlocs)
        tf1 = grid_finder.DictFormatter(dict(zip(tlocs, 
                                        map(str, rlocs))))
    
        std_min = std_range[0] * ref_std.min().item()
        std_max = std_range[1] * ref_std.max().item()

        ghelper = floating_axes.GridHelperCurveLinear(
                                tr,
                                extremes = (0, 
                                          tmax, 
                                          std_min, 
                                          std_max),
                                grid_locator1 = gl1, 
                                tick_formatter1 = tf1)

        markers_l = {10: 'o', 20: 'o', 30: 'o', 40: 'o'}

        markers_s = {k: 6 + scaler.transform([[float(k)]]) * 10 
                     for k in markers_l.keys()}
    
        style_1()
     
        nrows = 3
        ncols = 2
        w1, w2 = 1, 14

        fig = plt.figure(figsize=(6.7, 8), dpi = 300)

        gs = GridSpec(figure = fig, 
                      ncols = ncols, nrows = nrows, 
                      height_ratios = [w1, w2, w2])
        
        
        axl_l = fig.add_subplot(gs[0, 0], frameon = False)
        axl_r = fig.add_subplot(gs[0, 1], frameon = False)

        marker_legend(axl_l, 
                      markers_l, 
                      marker_color = 'gray',
                      marker_sizes = markers_s,
                      title = 'absolute PBIAS [%]',
                      anchor = (0.5, 0.5), 
                      markerfirst = True, 
                      fs_labels = 10,
                      labelcolor = 'k',
                      handletextpad = 0.2,
                      columnspacing = 0.9, 
                      loc = 'center',
                      handlelength = 1.2)
    
        color_legend(axl_r, 
                     colorscc,
                     cmapc, 
                     ncols = 2,
                     **color_legend_args)
        
        i_ax = [[1, 0], [1, 1], [2, 0], [2, 1]]

        order = ['ENF', 'DBF', 'GRA', 'CRO']

        lcs_df = std_df.index.unique(level = 'Landcover')

        lcs = [lc for lc in order if lc in lcs_df]
        
        for ilc, lc in enumerate(lcs):
            
            ref_std_lc = ref_std.loc[(lc, obs)].item()
            r_lc = r_df.loc[lc]
            std_lc = std_df.loc[lc]
            pbias_lc = pbias_df.loc[lc]
            ms_lc = pbias_ms.loc[lc]
            marker = markers[lc]

            ax = fig.add_subplot(gs[*i_ax[ilc]], 
                                 axes_class = floating_axes.FloatingAxes,
                                 grid_helper = ghelper,
                                 frameon = False)
        
            axgrid(ax)
                                     
            # Adjust axes
            ax.axis["top"].set_axis_direction("bottom")   # "Angle axis"
            ax.axis["top"].toggle(ticklabels = True, label=True)
            ax.axis["top"].major_ticklabels.set_axis_direction("top")
            ax.axis["top"].label.set_axis_direction("top")
            ax.axis["top"].label.set_text("Correlation [-]")

            ax.axis["left"].set_axis_direction("bottom")  # "X axis"
            ax.axis["left"].label.set_text(f"Standard deviation [{var_units}]")

            ax.axis["right"].set_axis_direction("top")    # "Y-axis"
            ax.axis["right"].toggle(ticklabels = True)
            ax.axis["right"].major_ticklabels.set_axis_direction(
                "bottom" if extend else "left")

            if std_min:
                ax.axis["bottom"].toggle(ticklabels=False, label=False)
            else:
                ax.axis["bottom"].set_visible(False)            

            ax = ax.get_aux_axes(tr)  
        
            #l_, = ax.plot([0], ref_std_lc, marker = marker,
            #             color = 'k', ls = '', ms = 10, label = '',
            #             clip_on = False, alpha = 0.7)
            t_ = np.linspace(0, tmax)
            r_ = np.zeros_like(t_) + ref_std_lc
            ax.plot(t_, r_, 'k--', alpha = 0.7, label = '_')

            colorsc = {k: cmapc[d] 
                       for k, d in colors.items()}
            
            rs, ts = np.meshgrid(np.linspace(std_min, std_max),
                             np.linspace(0, tmax))
            
            rms = np.sqrt(ref_std_lc**2 + 
                          rs ** 2 - 2 * ref_std_lc *
                          rs * np.cos(ts))
            
            contours = ax.contour(ts, 
                                  rs, 
                                  rms, 
                                  4, 
                                  colors = 'k', 
                                  linestyles = '--',
                                  linewidths = 0.6, 
                                  alpha = 0.5)
            
            ax.clabel(contours, 
                      inline = True, 
                      fmt = f'%.1f {var_units}',
                      fontsize = 8)

            props = dict(facecolor = 'white', 
                         edgecolor = 'none', 
                         alpha = 0.85)

            ax.text(0.5, 
                    1.0, 
                    lc, 
                    fontsize = 12, 
                    transform = ax.transAxes, 
                    va = 'bottom', 
                    ha = 'center', 
                    bbox = props)
            
            for src in std_df.index.get_level_values('Source'):

                r_src_lc = r_lc.loc[src].item()
                std_src_lc = std_lc.loc[src].item()
                pbias_src_lc = pbias_lc.loc[src].item()
                ms_src_lc = ms_lc.loc[src].item()

                

                if src == obs:
                    ms_src_lc = 5
                
                color = colorsc[src]
                
                ax.plot(np.arccos(r_src_lc), 
                        std_src_lc, 
                        marker = marker, 
                        color = color, 
                        markerfacecolor = "None",
                        markeredgewidth = 2.4,
                        ls = '', 
                        ms = ms_src_lc, 
                        label = '',
                        alpha = 0.8,
                        clip_on = False)

    return fig

def xy_landcover_moments(df: pd.DataFrame, 
                         variable: str, 
                         sources_insitu: list = [], 
                         sources_grids: list = [], 
                         sel_landcover: list = [], 
                         markers: dict = {}, 
                         colors: dict = {},
                         xy_init_args: dict = {}, 
                         xy_args: dict = {}, 
                         marker_legend_args: dict = {}, 
                         color_legend_args: dict = {},
                         moments: list = ['mean', 
                                          'variance', 
                                          'skewness', 
                                          'kurtosis']) -> plt.Figure:

    from my_.figures.two_by_two import square_two_top_cax
    from my_.series.group import select_multi_index, nona_level
    from my_.plot.init_ax import init_xy
    from my_.plot.basic import scatter, error
    from my_.plot.legend import marker_legend, color_legend
    #from my_.series.convert import tile_df_to_list

    from my_.resources.sources import query_variables

    from my_.math import stats

    import colorcet as cc
    import numpy as np

    print('Plot land cover aggregated xy plots...\n')

    stations_uq = df.columns.unique(level = 'Station')
    
    stations_count = df.columns\
                       .get_level_values('Station')\
                       .value_counts()
    
    
    for s in stations_uq:
        
        if stations_count[s] <= len(sources_grids):

            df = df.drop(s, axis = 1, level = 'Station')
            
    df_nona = nona_level(df, ['Variable', 'Station'], 
                         how = 'any', 
                         axis = 1)
    

    df_i = select_multi_index(df_nona, 
                              levels = ['Source', 
                                        'Variable', 
                                        'Landcover'],
                              keys = [sources_insitu, 
                                      [variable], 
                                      sel_landcover])
    

    df_d = select_multi_index(df_nona, 
                              levels = ['Source', 
                                        'Variable', 
                                        'Landcover'],
                              keys = [sources_grids, 
                                      [variable], 
                                      sel_landcover])
    
    df_i_lc = df_i.groupby(axis = 1, 
                           level = ['Source', 'Landcover'])
    
    df_d_lc = df_d.groupby(axis = 1, 
                           level = ['Source', 'Landcover'])

    i_sources = df_i_lc.obj.columns.unique(level = 'Source')

    d_sources = df_d_lc.obj.columns.unique(level = 'Source')

    cmapc = cc.glasbey_hv[:] + [[0.0, 0.0, 0.0]]

    var_units = query_variables(i_sources[0], 
                                'var_units')[variable]

    mom_units = [f'[{var_units}]', 
                 f'[{var_units}]²', '[-]', '[-]']

    if len(i_sources) > 1: raise NotImplementedError

    fig, axs, axs_l = square_two_top_cax(fy = 8)

    labels_sources = {s: query_variables(s, 'name_label') 
                      for s in d_sources}

    colors = {labels_sources[k]: v 
              for k,v in colors.items() if k in d_sources}
    


    marker_legend(axs_l[0], 
                  markers, 
                  **marker_legend_args)
    
    color_legend(axs_l[1], 
                 colors,
                 cmapc, 
                 ncols = (len(d_sources) // 2) + 1, 
                 **color_legend_args)

    for iax, ax in enumerate(axs):

        agg_moment = moments[iax]

        agg_method = getattr(stats, agg_moment)
        err_method = getattr(stats, f'std_error_{agg_moment}')
        
        df_i_lc_agg = df_i_lc.apply(agg_method)
        df_d_lc_agg = df_d_lc.apply(agg_method)

        df_i_lc_agg_err = df_i_lc.apply(err_method)
        df_d_lc_agg_err = df_d_lc.apply(err_method)

        title_fig = ''

        xlabel = f'Observation {mom_units[iax]}'
        ylabel = f'Model {mom_units[iax]}'

        init_xy(ax = ax, 
                xs = df_i_lc_agg, 
                ys = df_d_lc_agg,
                xerr = df_i_lc_agg_err,
                yerr = df_d_lc_agg_err,
                title = title_fig, 
                xlabel = xlabel, 
                ylabel = ylabel, 
                ax_tag = agg_moment,
                **xy_init_args)

        for lc in df_i.columns.unique(level = 'Landcover'):

            ys = select_multi_index(df_d_lc_agg, 
                                    'Landcover', 
                                    lc, 
                                    axis = 0)
            
            colorsc = [cmapc[colors[labels_sources[d]]] 
                    for d in ys\
                        .index\
                        .get_level_values('Source')]

            xs = select_multi_index(df_i_lc_agg, 
                                    'Landcover', 
                                    lc, 
                                    axis = 0).values.flatten()

            x_err = select_multi_index(df_i_lc_agg_err, 
                                       'Landcover', 
                                       lc, 
                                       axis = 0).values.flatten()

            y_err = select_multi_index(df_d_lc_agg_err, 
                                       'Landcover', 
                                       lc, 
                                       axis = 0).values.flatten()
            
            xs_tile = np.tile(xs, len(d_sources))
            
            x_err_tile = np.tile(x_err, len(d_sources))

            if len(ys) == 0: continue

            scatter(ax, 
                    xs_tile, 
                    ys.values.flatten(), markers[lc], 
                    colors_marker = colorsc, 
                    **xy_args)
            
            print(x_err_tile, y_err)

            error(ax, 
                  xs_tile, 
                  ys.values.flatten(), 
                  x_err = x_err_tile, 
                  y_err = y_err, 
                  ecolors = colorsc, 
                  elinewidth = 2)
    
    return fig


def bar_rmse_landcover(df, variable, sources_insitu, sources_grids, sel_landcover,
                       colors: dict = {}, bar_init_args = {}, bar_args = {},
                       color_legend_args = {}):

    from my_.series.group import select_multi_index, nona_level
    from my_.series.aggregate import column_wise
    from my_.math.stats import rmse, mean, std_deviation
    from my_.resources.sources import query_variables

    from my_.figures.two_by_two import square_top_cax
    from my_.plot.basic import bar, error
    from my_.plot.init_ax import init_bar
    from my_.plot.legend import color_legend

    import numpy as np
    import colorcet as cc

    print('Plot land cover aggregated rmse bar plots...\n')

    df_nona                     = nona_level(df, ['Variable', 'Station'],
                                             axis = 1,
                                             how = 'any')

    df_s                        = select_multi_index(df_nona, ['Variable', 'Landcover'],
                                                      keys = [variable, sel_landcover])
    
    df_s                        = df_s.reindex(labels = sources_insitu + sources_grids, axis = 1, level = 'Source')
    
    df_lc_rmse                  = df_s.groupby(axis = 1, level = ['Source', 'Landcover']).apply(column_wise, ffunc = rmse)

    df_lc_rmse_mean             = df_lc_rmse.groupby(axis = 1, level = ['Source', 'Landcover']).apply(mean, axis = (0, 1))
    df_lc_rmse_std             = df_lc_rmse.groupby(axis = 1, level = ['Source', 'Landcover']).apply(std_deviation, axis = (0, 1))

    sources                     = df_lc_rmse_mean.index.unique(level = 'Source')

    var_units                   = query_variables(sources[0], 'var_units')[variable]
    labels_sources              = {s: query_variables(s, 'name_label') for s in sources}
    
    fig, axs, axs_l             = square_top_cax()

    cmapc = cc.glasbey_hv[:] + [[0.0, 0.0, 0.0]]

    colors                      = {labels_sources[k]: v for k,v in colors.items() if k in sources}


    xs                          = np.arange(len(sources))

    ys                          = df_lc_rmse_mean

    ys = ys.reindex(sources_insitu + sources_grids, axis = 1, level = 'Source')

    color_legend(axs_l, colors, cmapc, **color_legend_args)

    y_err                       = df_lc_rmse_std

    y_err = y_err.reindex(sources_insitu + sources_grids, axis = 1, level = 'Source')
    
    for iax, ax in enumerate(axs):

        lc                      = sel_landcover[iax]

        ys_lc                   = select_multi_index(ys, levels = ['Landcover'], keys = [lc], axis = 0)

        colorsc = [cmapc[colors[labels_sources[d]]] 
                for d in ys_lc\
                    .index\
                    .get_level_values('Source')]

        y_err_lc                = select_multi_index(y_err, levels = ['Landcover'], keys = [lc], axis = 0)

        ecolors                 = ['gray'] * len(ys_lc)

        init_bar(ax, xs, ys + y_err, ax_tag = lc, ylabel = f'RMSD [{var_units}]', **bar_init_args)

        bar(ax, xs, ys_lc.values.flatten(), color = colorsc, **bar_args)
        
        error(ax, xs, ys_lc.values, y_err = y_err_lc.values, ecolors = ecolors, 
              capsize = 4, alpha = 0.8, zorder = 6)
    
    return fig


def doy_doy_landcover(name: str, 
                      df: pd.DataFrame,
                      variable: str,
                      sources_insitu: list[str], 
                      sources_grids: list[str], 
                      sel_landcover: list[str],
                      colors: dict = {}, 
                      doy_init_args: dict = {}, 
                      doy_args: dict = {}, 
                      color_legend_args: dict = {}) -> plt.Figure:

    from my_.series.group import select_multi_index, nona_level
    from my_.series.aggregate import column_wise
    from my_.math.stats import pbias, rmse
    from my_.resources.sources import query_variables
    from my_.figures.four_by_two import vertical_top_cax
    from my_.figures.two_by_two import square_two_top_cax
    from my_.plot.init_ax import init_ts
    from my_.plot.legend import color_legend
    from my_.plot.basic import plot
    from my_.files.handy import save_df 
    from scipy.signal import find_peaks
    from scipy.ndimage import gaussian_filter1d

    import colorcet as cc

    print('Plot land cover aggregated DOY and distribution plots...\n')

    stations_uq = df.columns.unique(level = 'Station')
    
    stations_count = df.columns\
                       .get_level_values('Station')\
                       .value_counts()
    
    for s in stations_uq:
        
        if stations_count[s] == 1:

            df = df.drop(s, axis = 1, level = 'Station')
    
    df_nona = nona_level(df, 
                         ['Variable', 'Station'],
                         axis = 1,
                         how = 'any')

    df_nona = df_nona.dropna(axis = 1, 
                             how = 'all')


    df_s = select_multi_index(df_nona, 
                              ['Variable', 'Landcover'],
                              keys = [variable, sel_landcover])
 
    df_s = df_s.reindex(labels = sources_insitu + sources_grids, 
                        axis = 1, 
                        level = 'Source')

    sources = df_s.columns.unique(level = 'Source')

    df_doy = df_s.groupby(df_s.index.dayofyear).mean()

    df_doy_lc = df_doy.groupby(axis = 1, level = ['Source', 
                                          'Landcover'])

    df_doy_lc_mean = df_doy_lc.mean()

    modes = df_doy_lc_mean.apply(find_peaks, distance = 46)

    df_doy_lc_mean_d2 = df_doy_lc_mean.apply(gaussian_filter1d, sigma = 100).apply(np.gradient).apply(np.gradient)

    df_doy_lc_mean_infl = df_doy_lc_mean_d2.apply(np.sign).apply(np.diff).apply(np.nonzero)

    df_doy_lc_mean_rmse = column_wise(df_doy_lc_mean, 
                                      ffunc = rmse)
    
    df_doy_lc_mean_pbias = column_wise(df_doy_lc_mean, 
                                       ffunc = pbias)

    save_df(df_doy_lc_mean, 
            f'out/{name}/csv/doy_lc_mean_{variable}.csv',
            format = 'csv')

    save_df(df_doy_lc_mean_rmse, 
            f'out/{name}/csv/doy_lc_mean_rmse_{variable}.csv', 
            format = 'csv')
    
    save_df(df_doy_lc_mean_pbias, 
            f'out/{name}/csv/doy_lc_mean_pbias_{variable}.csv', 
            format = 'csv')

    df_doy_lc_std = df_doy_lc.std()

    var_units = query_variables(sources[0], 
                                'var_units')[variable]
    
    labels_sources = {s: query_variables(s, 'name_label') 
                      for s in sources}

    cmapc = cc.glasbey_hv[:] + [[0.0, 0.0, 0.0]]
    
    colors = {labels_sources[k]: v 
              for k, v in colors.items() if k in sources}

    fig, axs, axs_l = vertical_top_cax(fy = 8)

    color_legend(axs_l, colors, cmapc, **color_legend_args)


    for ilc, lc in enumerate(sel_landcover):

        xs_doy = df_doy_lc_mean.index

        ys_doy = select_multi_index(df_doy_lc_mean, 
                                    levels = ['Landcover'], 
                                    keys = [lc])
        
        if len(ys_doy.columns) < 2: continue
        
        ys_doy = ys_doy.reindex(labels = sources_insitu + sources_grids, 
                        axis = 1, 
                        level = 'Source')
        
        colorsc = [cmapc[colors[labels_sources[d]]] 
                    for d in ys_doy\
                        .columns\
                        .get_level_values('Source')]
        
        err_doy = select_multi_index(df_doy_lc_std, 
                                     levels = ['Landcover'], 
                                     keys = [lc])
        
        err_doy = err_doy.reindex(labels = sources_insitu + sources_grids, 
                        axis = 1, 
                        level = 'Source')

        var_label = f'{variable} [{var_units}]'

        init_ts(ax = axs[ilc, 0], 
                xs = np.arange(1, 365, 8), 
                ys = df_doy_lc_mean, 
                ylabel = f'mean {var_label}', 
                ax_tag = lc, 
                **doy_init_args)
        
        init_ts(ax = axs[ilc, 1], 
                xs = np.arange(1, 365, 8), 
                ys = df_doy_lc_std, 
                ylabel = f'std {var_label}', 
                ax_tag = lc, 
                **doy_init_args)

        plot(ax = axs[ilc, 0], 
             xs = xs_doy, 
             ys = ys_doy,
             lw = [3] + [1.5] * len(sources_grids), 
             colors = colorsc, 
             **doy_args)
        
        plot(ax = axs[ilc, 1], 
             xs = xs_doy, 
             ys = err_doy, 
             lw = [3] + [1.5] * len(sources_grids),
             colors = colorsc, 
             **doy_args)

    return fig


def knee(xy: pd.Series, 
        interp_method = 'interp1d',
        degree: int = 7,
        curve: str = 'convex',
        direction: str = 'increasing'):
    
    from kneed import KneeLocator

    kneeds = KneeLocator(x = xy.index,
                         y = xy.values,
                         polynomial_degree = degree,
                         interp_method = interp_method,
                         curve = curve,
                         direction = direction)
    
    return kneeds.knee



def doy_lollipop(df: pd.DataFrame,
                 variable: str,
                 sel_landcover: list[str],
                 sources_insitu: list[str],
                 sources_grids: list[str],
                 color_legend_args: dict,
                 colors: dict):

    from my_.series.group import select_multi_index, nona_level
    from my_.resources.sources import query_variables
    from my_.figures.two_by_two import square_top_cax
    from my_.plot.init_ax import init_lol
    from my_.plot.legend import color_legend
    from my_.plot.basic import scatter, hlines, vlines, error
    from my_.files.handy import save_df 
    from my_.math.stats import peaks, inflictions
    from scipy.ndimage import gaussian_filter1d
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    


    import colorcet as cc

    import string

        
    abc                     = string.ascii_lowercase
    
    list_abc                = list(abc)

    print('Plot land cover aggregated DOY lollipop...\n')

    stations_uq = df.columns.unique(level = 'Station')
    
    stations_count = df.columns\
                       .get_level_values('Station')\
                       .value_counts()
    
    for s in stations_uq:
        
        if stations_count[s] == 1:

            df = df.drop(s, axis = 1, level = 'Station')
    
    df_nona = nona_level(df, 
                         ['Variable', 'Station'],
                         axis = 1,
                         how = 'any')

    df_nona = df_nona.dropna(axis = 1, 
                             how = 'all')

    df_s = select_multi_index(df_nona, 
                              ['Variable', 'Landcover'],
                              keys = [variable, sel_landcover])
 
    df_s = df_s.reindex(labels = sources_insitu + sources_grids, 
                        axis = 1, 
                        level = 'Source')

    sources = df_s.columns.unique(level = 'Source')

    df_doy = df_s.groupby(df_s.index.dayofyear).mean()

    df_doy_modes = df_doy.apply(peaks, distance = 46)

    df_doy_modes_lc_mean = df_doy_modes.groupby(level = ['Source', 
                                                'Landcover']).mean()
    
    df_doy_modes_lc_mean = pd.DataFrame(df_doy_modes_lc_mean).T
    
    df_doy_modes_lc_std = df_doy_modes.groupby(level = ['Source', 
                                               'Landcover']).std()
    
    df_doy_modes_lc_std = pd.DataFrame(df_doy_modes_lc_std).T

    df_doy_infl = df_doy.apply(inflictions)

    df_doy_infl_lc_mean = df_doy_infl.groupby(axis = 1,
                                              level = ['Source', 
                                              'Landcover']).mean()
    
    df_doy_infl_lc_std = df_doy_infl.groupby(axis = 1,
                                             level = ['Source', 
                                             'Landcover']).std()
    
    save_df(df_doy_infl_lc_mean, f'infl_mean_{variable}.csv')
    save_df(df_doy_infl_lc_std, f'infl_std_{variable}.csv')
    save_df(df_doy_modes_lc_mean, f'modes_mean_{variable}.csv')
    save_df(df_doy_modes_lc_std, f'modes_std_{variable}.csv')

    labels_sources = {s: query_variables(s, 'name_label') 
                      for s in sources}

    cmapc = cc.glasbey_hv[:] + [[0.0, 0.0, 0.0]]
    
    colors = {labels_sources[k]: v 
              for k, v in colors.items() if k in sources_grids}

    fig, axs, axs_l = square_top_cax(fx = 9.4,
                                     annotation = False)

    color_legend(axs_l, 
                 colors, 
                 cmapc, 
                 **color_legend_args)


    for ilc, lc in enumerate(sel_landcover):

        xs_mode = select_multi_index(df_doy_modes_lc_mean, 
                                    levels = ['Landcover'], 
                                    keys = [lc])
        
        err_mode = select_multi_index(df_doy_modes_lc_std, 
                                    levels = ['Landcover'], 
                                    keys = [lc])
        
        xs_infl = select_multi_index(df_doy_infl_lc_mean, 
                                    levels = ['Landcover'], 
                                    keys = [lc])
        
        err_infl = select_multi_index(df_doy_infl_lc_std, 
                                      levels = ['Landcover'], 
                                      keys = [lc])
        
        ys_source = np.arange(len(sources_grids))

        ys_obs = np.arange(len(sources_grids))
        
        if len(xs_mode.columns) < 2: continue
        
        xmin = select_multi_index(xs_mode,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_insitu[0],
                                          lc]).T * 8
        
        xmin_err = select_multi_index(err_mode,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_insitu[0],
                                          lc]).T * 8

        xmax = select_multi_index(xs_mode,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_grids,
                                          lc]).T * 8

        xmax_err = select_multi_index(err_mode,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_grids,
                                          lc]).T * 8
        
        xinfl = select_multi_index(xs_infl,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_grids,
                                          lc]).T * 8
        
        err_xinfl = select_multi_index(err_infl,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_grids,
                                          lc]).T * 8
        
        xinfl_o = select_multi_index(xs_infl,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_insitu[0],
                                          lc]).T * 8
        
        err_xinfl_o = select_multi_index(err_infl,
                                  levels = ['Source',
                                            'Landcover'],
                                  keys = [sources_insitu[0],
                                          lc]).T * 8
        
        xmax = xmax.reindex(sources_grids, 
                            axis = 0,
                            level = 'Source')
        
        xmax_err = xmax_err.reindex(sources_grids, 
                                    axis = 0,
                                     level = 'Source')
        
        xinfl = xinfl.reindex(sources_grids, 
                                     axis = 0, 
                                     level = 'Source')

        err_xinfl = err_xinfl.reindex(sources_grids, 
                                      axis = 0,
                                     level = 'Source')
        
        colorsc = [cmapc[colors[labels_sources[d]]] 
                    for d in xmax\
                        .index\
                        .get_level_values('Source')]
        
        init_lol(ax = axs[ilc], 
                 xs = xmax, 
                 ys = ys_source,
                 title = lc,
                 ax_tag = 'Peak', 
                 axhv_ls = '--',
                 axhv_lw = 1,
                 axhv_dashes = (4, 4),
                 fs_title = 14,
                 y_title = 1.05,
                 fs_label = 10,
                 fs_ticks = 10,
                 ax_tag_x = 0.5,
                 ax_tag_y = 1,
                 ylabel = '',
                 xlabel = 'Day of year')
        
        hlines(axs[ilc],
               ys = ys_source,
               xmin = xmin,
               xmax = xmax,
               colors = colorsc,
               linewidth = 7,
               alpha = 0.9)
        
        vlines(axs[ilc],
               xs = xmin.values,
               ymin = -1,
               ymax = 0.95,
               colors = 'k',
               linewidth = 3,
               linestyle = 'solid',
               alpha = 0.7)
        
        axs[ilc].vlines(
               x = [0, 1],
               ymin = -1,
               ymax = 0.95,
               colors = 'k',
               linewidth = 2,
               linestyle = 'dashed',
               alpha = 0.7,
               transform = axs[ilc].transAxes)
        
        scatter(axs[ilc],
                xs = xmax.values,
                ys = ys_source,
                colors_marker = colorsc,
                marker = 'o',
                sizes_marker = 150,
                alpha = 0.9)

        error(axs[ilc],
              xs = xmax.values,
              ys = ys_source,
              x_err = xmax_err.values,
              ecolors = ['darkgray'] * len(ys_source),
              elinewidth = 2.5,
              capsize = 6, 
              alpha = 0.9, 
              zorder = 6)
        
        error(axs[ilc],
              xs = xmin.values,
              ys = [-0.35],
              x_err = xmin_err.values,
              ecolors = ['k'],
              capsize = 6,
              elinewidth = 2.5, 
              alpha = 0.8, 
              zorder = 6)

        axs[ilc].set_ylim((-0.5, len(sources_grids)))
        axs[ilc].set_xlim((np.min(np.concatenate([xmax.values - xmax_err.values, 
                         xmin.values - xmin_err.values])) - 16, 
                         np.max(np.concatenate([xmax.values + xmax_err.values, 
                         xmin.values + xmin_err.values])) + 16))
        axs[ilc].set_yticks([], labels =[])

        divider = make_axes_locatable(axs[ilc])
        axinfl1 = divider.append_axes("left", size="100%", pad=0.1, sharey=axs[ilc])
        axinfl2 = divider.append_axes("right", size="100%", pad=0.1, sharey=axs[ilc])

        init_lol(ax = axinfl1, 
                 xs = xinfl_o, 
                 ys = ys_source,
                 ax_tag = 'Season start', 
                 axhv_ls = '--',
                 axhv_lw = 1,
                 axhv_dashes = (4, 4),
                 fs_title = 14,
                 y_title = 1.05,
                 fs_label = 10,
                 fs_ticks = 10,
                 ax_tag_x = 0.5,
                 ax_tag_y = 1,
                 ylabel = '',
                 xlabel = '')
        
        hlines(axinfl1,
               ys = ys_source,
               xmin = xinfl_o.iloc[:, 0],
               xmax = xinfl.iloc[:, 0],
               colors = colorsc,
               linewidth = 7,
               alpha = 0.9)
        
        vlines(axinfl1,
               xs = xinfl_o.iloc[:, 0].values,
               ymin = -1,
               ymax = 0.95,
               colors = 'k',
               linewidth = 3,
               linestyle = 'solid',
               alpha = 0.8)
        
        axinfl1.vlines(
               x = [1],
               ymin = -1,
               ymax = 0.95,
               colors = 'k',
               linewidth = 2,
               linestyle = 'dashed',
               alpha = 0.7,
               transform = axinfl1.transAxes)
        
        scatter(axinfl1,
                xs = xinfl.iloc[:,0],
                ys = ys_source,
                colors_marker = colorsc,
                marker = 'o',
                sizes_marker = 150,
                alpha = 0.9)
        
        error(axinfl1,
              xs = xinfl.iloc[:, 0],
              ys = ys_source,
              x_err = err_xinfl.iloc[:, 0],
              ecolors = ['darkgray'] * len(ys_source),
              elinewidth = 2.5,
              capthick = 2,
              capsize = 6, 
              alpha = 0.9, 
              zorder = 6)
        
        error(axinfl1,
              xs = xinfl_o.iloc[:, 0],
              ys = [-0.35],
              x_err = err_xinfl_o.iloc[:, 0],
              ecolors = ['k'],
              capsize = 6,
              capthick = 2,
              elinewidth = 2.5, 
              alpha = 0.8, 
              zorder = 6)
        
        axinfl1.set_ylim((-0.5, len(sources_grids)))
        axinfl1.set_xlim(np.min(np.concatenate([xinfl.iloc[:, 0].values - err_xinfl.iloc[:, 0].values, 
                         xinfl_o.iloc[:, 0].values - err_xinfl_o.iloc[:, 0].values])) - 16, 
                         np.max(np.concatenate([xinfl.iloc[:, 0].values + err_xinfl.iloc[:, 0].values, 
                         xinfl_o.iloc[:, 0].values + err_xinfl_o.iloc[:, 0].values])) + 16)
        axinfl1.set_yticklabels([])

        axinfl1.text(0.5, 1.10, list_abc[ilc] + ')', fontsize = 14,
                transform = axinfl1.transAxes,
                va = 'bottom', 
                ha = 'center')

        init_lol(ax = axinfl2, 
                 xs = xinfl_o, 
                 ys = ys_source,
                 ax_tag = 'Season end', 
                 axhv_ls = '--',
                 axhv_lw = 1,
                 axhv_dashes = (4, 4),
                 fs_title = 14,
                 y_title = 1.05,
                 fs_label = 10,
                 fs_ticks = 10,
                 ax_tag_x = 0.5,
                 ax_tag_y = 1,
                 ylabel = '',
                 xlabel = '')
        
        hlines(axinfl2,
               ys = ys_source,
               xmin = xinfl_o.iloc[:, 1],
               xmax = xinfl.iloc[:, 1],
               colors = colorsc,
               linewidth = 7,
               alpha = 0.9)
        
        vlines(axinfl2,
               xs = xinfl_o.iloc[:, 1].values,
               ymin = -1,
               ymax = 0.95,
               colors = 'k',
               linewidth = 3,
               linestyle = 'solid',
               alpha = 0.7)
        
        axinfl2.vlines(
               x = [0],
               ymin = -1,
               ymax = 0.95,
               colors = 'k',
               linewidth = 2,
               linestyle = 'dashed',
               alpha = 0.8,
               transform = axinfl2.transAxes)
        
        scatter(axinfl2,
                xs = xinfl.iloc[:, 1],
                ys = ys_source,
                colors_marker = colorsc,
                marker = 'o',
                sizes_marker = 150,
                alpha = 0.9)
        
        error(axinfl2,
              xs = xinfl.iloc[:, 1],
              ys = ys_source,
              x_err = err_xinfl.iloc[:, 1],
              ecolors = ['darkgray'] * len(ys_source),
              elinewidth = 2.5,
              capthick = 2,
              capsize = 6, 
              alpha = 0.9, 
              zorder = 6)
        
        error(axinfl2,
              xs = xinfl_o.iloc[:, 1],
              ys = [-0.35],
              x_err = err_xinfl_o.iloc[:, 1],
              ecolors = ['k'],
              capsize = 6,
              capthick = 2,
              elinewidth = 2.5, 
              alpha = 0.8, 
              zorder = 6)
        
        axinfl2.set_ylim((-0.5, len(sources_grids)))
        axinfl2.set_xlim(np.min(np.concatenate([xinfl.iloc[:, 1].values - err_xinfl.iloc[:, 1].values, 
                         xinfl_o.iloc[:, 1].values - err_xinfl_o.iloc[:, 1].values])) - 16, 
                         np.max(np.concatenate([xinfl.iloc[:, 1].values + err_xinfl.iloc[:, 1].values, 
                         xinfl_o.iloc[:, 1].values + err_xinfl_o.iloc[:, 1].values])) + 16)
        axinfl2.set_yticklabels([])

    return fig

def doy_dist_landcover(name, df, variable, sources_insitu, sources_grids, sel_landcover,
                       colors: dict = {}, doy_init_args = {}, dist_init_args = {},
                       doy_args = {}, dist_args = {}, doy_fill_args = {},
                       color_legend_args = {}, do_fill = True, do_line_bounds = False):


    from my_.series.group import select_multi_index, nona_level
    from my_.series.aggregate import column_wise
    from my_.math.stats import gauss_kde_pdf, pbias, rmse
    from my_.resources.sources import query_variables
    from my_.figures.four_by_two import vertical_top_cax
    from my_.plot.init_ax import init_ts, init_dist
    from my_.plot.legend import color_legend
    from my_.plot.basic import plot, fill 
    from my_.files.handy import save_df 

    import colorcet as cc

    print('Plot land cover aggregated DOY and distribution plots...\n')
    
    df_nona                     = nona_level(df,
                                             ['Variable', 'Station'],
                                             axis = 1,
                                             how = 'any')

    df_s                        = select_multi_index(df_nona, ['Variable', 'Landcover'],
                                                      keys = [variable, sel_landcover])
    
    df_s                        = df_s.reindex(labels = sources_insitu + sources_grids, axis = 1, level = 'Source')

    sources                     = df_s.columns.unique(level = 'Source')

    df_doy                      = df_s.groupby(df_s.index.dayofyear).mean()

    df_doy_lc                   = df_doy.groupby(axis = 1, level = ['Source', 'Landcover'])

    df_doy_lc_mean              = df_doy_lc.mean()

    df_doy_lc_mean_rmse         = column_wise(df_doy_lc_mean, ffunc = rmse)
    df_doy_lc_mean_pbias        = column_wise(df_doy_lc_mean, ffunc = pbias)
    save_df(df_doy_lc_mean_rmse, f'out/{name}/csv/doy_lc_mean_rmse_{variable}.csv', format = 'csv')
    save_df(df_doy_lc_mean_pbias, f'out/{name}/csv/doy_lc_mean_pbias_{variable}.csv', format = 'csv')

    df_doy_lc_std               = df_doy_lc.std()

    df_dist_lc                  = df_s.groupby(axis = 1, level = ['Source', 'Landcover']).apply(gauss_kde_pdf, return_dict=True)

    df_dist_lc.columns          = df_dist_lc.columns.set_names('pdf', level = -1)

    all_xs                      = select_multi_index(df_dist_lc, 'pdf', 'xs')
    all_ys                      = select_multi_index(df_dist_lc, 'pdf', 'ys')

    fig, axs, axs_l             = vertical_top_cax(fy = 8)

    var_units                   = query_variables(sources[0], 'var_units')[variable]
    labels_sources              = {s: query_variables(s, 'name_label') for s in sources}

    cmapc                       = cc.glasbey_hv[:]
    colors                      = {labels_sources[k]: v for k,v in colors.items() if k in sources}
    colorsc                     = [cmapc[colors[d]] for d in colors.keys()]

    color_legend(axs_l, colors, cmapc, **color_legend_args)

    for ilc, lc in enumerate(sel_landcover):

        xs_doy                  = df_doy_lc_mean.index

        ys_doy                  = select_multi_index(df_doy_lc_mean, levels = ['Landcover'], keys = [lc])
        
        err_doy                 = select_multi_index(df_doy_lc_std, levels = ['Landcover'], keys = [lc])

        xs_dist                 = select_multi_index(df_dist_lc, levels = ['Landcover', 'pdf'], keys = [lc, 'xs'])

        ys_dist                 = select_multi_index(df_dist_lc, levels = ['Landcover', 'pdf'], keys = [lc, 'ys'])

        var_label               = f'{variable} [{var_units}]'

        lower                   = ys_doy - err_doy
        upper                   = ys_doy + err_doy

        init_ts(axs[ilc, 0], xs_doy, df_doy_lc_mean, ylabel = var_label, ax_tag = lc, **doy_init_args)
        init_dist(axs[ilc, 1], all_xs, all_ys, xlabel = var_label, ax_tag = lc, **dist_init_args)

        plot(axs[ilc, 0], xs_doy, ys_doy, colors = colorsc, **doy_args)
        if do_fill : fill(axs[ilc, 0], xs_doy, lower, upper, colors = colorsc, **doy_fill_args)

        if do_line_bounds:
            plot(axs[ilc, 0], xs_doy, lower, colors = colorsc, style = '-.', alpha = 0.6, lw = 1.0, zorder = 6)
            plot(axs[ilc, 0], xs_doy, upper, colors = colorsc, style = '-.', alpha = 0.6, lw = 1.0, zorder = 6)
        plot(axs[ilc, 1], xs_dist, ys_dist, colors = colorsc, **dist_args)
        
    return fig

def dist_landcover(name: str, 
                   df: pd.DataFrame, 
                   variable: str, 
                   sources_insitu: list[str], 
                   sources_grids: list[str], 
                   sel_landcover: list[str],
                   colors: dict = {}, 
                   dist_init_args: dict = {},
                   dist_args: dict = {},
                   color_legend_args: dict = {}):
    
    from my_.series.group import select_multi_index, nona_level
    from my_.series.aggregate import column_wise
    from my_.math.stats import gauss_kde_pdf, pbias, rmse
    from my_.resources.sources import query_variables
    from my_.figures.two_by_two import square_top_cax
    from my_.plot.init_ax import init_ts, init_dist
    from my_.plot.legend import color_legend
    from my_.plot.basic import plot, fill 
    from my_.files.handy import save_df

    import colorcet as cc

    print('Plot land cover aggregated DOY and distribution plots...\n')
    
    df_nona                     = nona_level(df,
                                             ['Variable', 'Station'],
                                             axis = 1,
                                             how = 'any')

    df_s                        = select_multi_index(df_nona, ['Variable', 'Landcover'],
                                                      keys = [variable, sel_landcover])
    
    #df_s                        = df_s.reindex(sources_insitu + sources_grids, axis = 1, level = 'Source')

    sources                     = df_s.columns.unique(level = 'Source')

    df_dist_lc                  = df_s.groupby(axis = 1, level = ['Source', 'Landcover']).apply(gauss_kde_pdf, return_dict=True)


    df_dist_lc.columns          = df_dist_lc.columns.set_names('pdf', level = -1)

    all_xs                      = select_multi_index(df_dist_lc, 'pdf', 'xs')
    all_ys                      = select_multi_index(df_dist_lc, 'pdf', 'ys')

    var_units                   = query_variables(sources[0], 'var_units')[variable]
    labels_sources              = {s: query_variables(s, 'name_label') for s in sources}

    cmapc                       = cc.glasbey_hv[:] + [[0.0, 0.0, 0.0]]
    colorsp                     = {labels_sources[k]: v for k, v in colors.items() if k in sources}

    fig, axs, axl               = square_top_cax(fy = 5.5)

    color_legend(axl, colorsp, cmapc, **color_legend_args)

    for ilc, lc in enumerate(sel_landcover):

        lw = [3] + [1.5] * len(sources_grids)

        xs_dist = select_multi_index(df_dist_lc, levels = ['Landcover', 'pdf'], keys = [lc, 'xs'])

        xs_dist = xs_dist.reindex(sources_insitu + sources_grids, axis = 1, level = 'Source')

        ys_dist = select_multi_index(df_dist_lc, levels = ['Landcover', 'pdf'], keys = [lc, 'ys'])

        ys_dist = ys_dist.reindex(sources_insitu + sources_grids, axis = 1, level = 'Source')

        colorsc = [cmapc[colorsp[labels_sources[d]]] 
                    for d in ys_dist\
                        .columns\
                        .get_level_values('Source')]

        var_label               = f'{variable} [{var_units}]'

        init_dist(axs[ilc], all_xs, all_ys, xlabel = var_label, ax_tag = lc, **dist_init_args)

        plot(axs[ilc], xs_dist, ys_dist, lw = lw, colors = colorsc, **dist_args)

    return fig

def pie_landcover(df_static, selected_landcover, colors = {}, color_legend_args = {}):

    print('Plot network land cover pie plots\n')

    from my_.series.aggregate import concat
    from my_.resources.sources import query_static
    from my_.figures.double import horizontal_top_cax
    from my_.plot.basic import pie
    from my_.plot.legend import color_legend
    from my_.plot.init_ax import init_pie
    import matplotlib.pyplot as plt

    landcover                   = df_static['landcover'].where(df_static['landcover'].isin(selected_landcover))

    pct_crop                    = df_static['PCT_CROP'] / 100

    pct_nat                     = df_static['PCT_NATVEG'] / 100

    pft_cols                    = [col for col in df_static.columns if 'PCT_NAT_PFT_' in col]

    cft_cols                    = [col for col in df_static.columns if 'PCT_CFT_' in col]

    pft_rel                     = df_static[pft_cols].multiply(pct_nat, axis = 0)

    cft_rel                     = df_static[cft_cols].multiply(pct_crop, axis = 0)

    df_tot                      = concat([pft_rel, cft_rel], sort = False)

    ind_landcover               = query_static('CLM5-EU3-surf', 'sel_agg_layer_PFT')

    shares_surf                 = [df_tot.iloc[:, ind_landcover[lc]].sum(axis = 1).mean() for lc in selected_landcover]

    shares_surf_o               = shares_surf + [100 - sum(shares_surf)] 

    shares_lc                   = landcover.value_counts(normalize = True)

    fig, axs, axl               = horizontal_top_cax(x_an = 0.1, y_an = 0.8)

    cmapc                       = plt.get_cmap('Set2').colors

    colors                      = {k: v for k,v in colors.items()}
    colorsc                     = [cmapc[colors[lc]] for lc in list(selected_landcover)]

    colors['other']             = 7

    colorsc_o                   = colorsc + [cmapc[7]]

    color_legend(axl, colors, cmapc, **color_legend_args)

    init_pie(axs[0], ax_tag = 'ICOS network')
    init_pie(axs[1], ax_tag = 'Corresponding CLM5 surface')

    pie(axs[0], shares_lc, colorsc, autopct='%1.1f%%')
    pie(axs[1], shares_surf_o, colorsc_o, autopct='%1.1f%%')

    return fig


def doy_landcover(df, variable, sources_insitu, sources_grids, sel_landcover,
                       stat: str = 'mean', stat_err: str = 'std',
                       colors: dict = {}, doy_init_args = {},
                       doy_args = {}, doy_fill_args = {},
                       color_legend_args = {}, do_fill = False):
    
    print('Plot network land cover DOY plots\n')

    from my_.series.group import select_multi_index, nona_level
    from my_.resources.sources import query_variables
    from my_.figures.two_by_two import square_top_cax
    from my_.plot.init_ax import init_ts
    from my_.plot.legend import color_legend
    from my_.plot.basic import plot, fill 

    import colorcet as cc

    df_nona                     = nona_level(df, 
                                             ['Variable', 'Station'],
                                             axis = 1,
                                             how = 'any')

    df_s                        = select_multi_index(df_nona, ['Variable', 'Landcover'],
                                                      keys = [variable, sel_landcover])
    
    df_s                        = df_s.reindex(labels = sources_insitu + sources_grids, axis = 1, level = 'Source')

    sources                     = df_s.columns.unique(level = 'Source')

    df_doy                      = df_s.groupby(df_s.index.dayofyear).mean()

    df_doy_lc                   = df_doy.groupby(axis = 1, level = ['Source', 'Landcover'])

    df_doy_lc_agg               = df_doy_lc.agg(stat)

    df_doy_lc_agg_err           = df_doy_lc.agg(stat_err)

    fig, axs, axl               = square_top_cax(fx = 8)

    var_units                   = query_variables(sources[0], 'var_units')[variable]
    labels_sources              = {s: query_variables(s, 'name_label') for s in sources}

    cmapc                       = cc.glasbey_hv[:]
    colors                      = {labels_sources[k]: v for k,v in colors.items() if k in sources}
    colorsc                     = [cmapc[colors[d]] for d in colors.keys()]

    color_legend(axl, colors, cmapc, **color_legend_args)

    for ilc, lc in enumerate(sel_landcover):

        xs_doy                  = df_doy_lc_agg.index

        ys_doy                  = select_multi_index(df_doy_lc_agg, levels = ['Landcover'], keys = [lc])
        
        err_doy                 = select_multi_index(df_doy_lc_agg_err, levels = ['Landcover'], keys = [lc])

        var_label               = f'{variable} [{var_units}]'

        lower                   = ys_doy - err_doy
        upper                   = ys_doy + err_doy

        init_ts(axs[ilc], xs_doy, df_doy_lc_agg, ylabel = var_label, ax_tag = lc, **doy_init_args)
        plot(axs[ilc], xs_doy, ys_doy, colors = colorsc, **doy_args)

        if do_fill: fill(axs[ilc], xs_doy, lower, upper, colors = colorsc, **doy_fill_args)

    return fig


def plot_ts(df, variable, station, colors: dict = {}, color_legend_args: dict = {},
            fig_args: dict = {}, ts_init_args: dict = {}, plot_args: dict = {}):

    print(f'Plot station time-series: {station}, {variable}\n')

    from my_.series.group import select_multi_index, nona_level
    from my_.figures.single import square_top_cax
    from my_.plot.init_ax import init_ts_2
    from my_.plot.basic import plot
    from my_.plot.legend import color_legend
    from my_.resources.sources import query_variables

    import colorcet as cc
    import numpy as np


    stations_uq = df.columns.unique(level = 'Station')
    
    stations_count = df.columns\
                       .get_level_values('Station')\
                       .value_counts()
    
    for s in stations_uq:
        
        if stations_count[s] == 1:

            df = df.drop(s, axis = 1, level = 'Station')

    df_nona                     = nona_level(df, 
                                             ['Variable', 'Station'],
                                             axis = 1,
                                             how = 'any')

    df_s                        = select_multi_index(df_nona, ['Variable', 'Station'],
                                                      keys = [variable, station])
    
    sources                     = df_s.columns.unique(level = 'Source')

    fig, ax, axl, gs            = square_top_cax(**fig_args)

    if len(sources) == 0 : return fig
    
    cmapc                       = cc.glasbey_hv[:] + [[0.0, 0.0, 0.0]]
    colors                      = {k: v for k,v in colors.items() if k in sources}
    colorsc                     = [cmapc[colors[d]] for d in list(sources)]

    var_units                   = query_variables(sources[0], 'var_units')[variable]

    xs                          = df_s.index
    ys                          = df_s

    if np.all(np.isnan(ys)): return fig
    
    color_legend(axl, colors, cmapc, **color_legend_args)

    init_ts_2(ax, df_s.dropna(how = 'all').index, df_s.dropna(how = 'all'),
            title = f'{station} {variable} [{var_units}]', **ts_init_args)
    
    plot(ax, xs, ys, colors = colorsc, cycle_direction = 1, **plot_args)

    return  fig


def plot_ts_2(x, y, labels, variable, colors: dict = {}, color_legend_args: dict = {},
            fig_args: dict = {}, ts_init_args: dict = {}, plot_args: dict = {}):

    print(f'Plot station time-series: {variable}\n')

    from my_.figures.single import square_top_cax
    from my_.plot.init_ax import init_ts_2
    from my_.plot.basic import plot
    from my_.plot.legend import color_legend
    from my_.resources.sources import query_variables

    import colorcet as cc
    import numpy as np

    cmapc                       = cc.glasbey_hv[:]
    colors                      = {k: v for k, v in colors.items() if k in labels}
    colorsc                     = [cmapc[colors[d]] for d in list(labels)]

    fig, ax, axl, gs            = square_top_cax(**fig_args)

    #var_units                   = query_variables('CLM5-EU3', 'var_units')[variable]

    if np.all(np.isnan(y)): return fig
    
    color_legend(axl, colors, cmapc, **color_legend_args)

    init_ts_2(ax, x, y, **ts_init_args)

    plot(ax, x, y, colors = colorsc, **plot_args)

    return  fig




def stat(array, method):

    import numpy as np

    import my_math.stats as stats

    func                = getattr(stats, method)

    array_mean          = func(array, axis = (-1, -2))

    return array_mean


def mask(array, mask, true_value):

    import numpy as np

    array_masked        = np.where(mask == true_value, array, np.nan)

    return array_masked



import pandas as pd
import numpy as np

def linear(values, multiplicant, addend):
    
    new_values = values * multiplicant + addend

    return new_values


def relative_difference(array1, array2, percent: bool = False):

    from my_.math.arithmetic import difference, division

    absolute_difference         = difference(array1, array2)

    relative_difference         = division(absolute_difference, array2)

    if percent == True: return relative_difference * 100

    return relative_difference





import pandas as pd

def open_csv(file_str: str,
             **args) -> pd.DataFrame:

    import pandas as pd
    from glob import glob

    file                       = glob(file_str)[0]

    df                          = pd.read_csv(file, **args)

    return df





def addition(augend, addend):

    sum                     = augend + addend

    return sum


def difference(minuend, subtrahend):

    diff                    = minuend - subtrahend 

    return diff


def multiplication(multiplier, multiplicand):

    product                 = multiplier * multiplicand

    return product


def division(numerator, denominator):
    
    quotient                = numerator / denominator

    return quotient


def nth_root(radicand, degree):

    root                    = radicand ** (division(1, degree))

    return root


def exponentiation(base, exponent):

    power                   = base ** exponent

    return power


def logarithm(logx, base):

    from math import log

    logarithm               = log(logx, base)

    return logarithm



import matplotlib.pyplot as plt

def colorbar(cax: plt.Axes, 
             artist: plt.Artist, 
             ylabel: str = '',
             pad: int = 20, 
             shrink = 0.5,
             fraction: float = 0.05,
             extend = 'neither', 
             fs_label: float = 10, 
             tick_labels: list = [],
             orientation: str = 'vertical',
             rotation: float = 270):

    import numpy as np
    
    cbar = plt.colorbar(artist, 
                        cax = cax, 
                        extend = extend, 
                        shrink = shrink, 
                        fraction = fraction,
                        orientation = orientation)
    
    if orientation == 'vertical':

        cbar.ax.set_ylabel(ylabel, 
                           labelpad = pad, 
                           rotation = rotation, 
                           fontsize = fs_label)
    
    elif orientation == 'horizontal': 

        cbar.ax.set_xlabel(ylabel, 
                           labelpad = pad, 
                           rotation = rotation, 
                           fontsize = fs_label)

    len_tick_labels         = len(tick_labels)

    if len_tick_labels > 0:

        positions           = np.linspace(0.5, 
                                          (len_tick_labels - 1) - 0.5, 
                                          len_tick_labels)

        if orientation == 'vertical':
        
            cbar.ax.set_yticks(positions, 
                               tick_labels, 
                               fontsize = fs_label)

        if orientation == 'horizontal':

            cbar.ax.set_xticks(positions, 
                               tick_labels, 
                               fontsize = fs_label)
            
    return cbar


def colormap(cmap: str = 'viridis', cmap_n = 1000):

    import matplotlib.pyplot as plt


    cmap_c                  = plt.get_cmap(cmap, cmap_n)

    return cmap_c


def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
  
    import matplotlib.colors as colors
    import numpy as np
  
    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    
    return new_cmap   



import numpy as np
import xarray as xr

def vpd_allen_2000(temp_max: np.ndarray | xr.DataArray,
                   temp_min: np.ndarray | xr.DataArray,
                   temp: np.ndarray | xr.DataArray,
                   rh: np.ndarray | xr.DataArray,
                   ) -> np.ndarray | xr.DataArray:
    
    # Temperature in ° Celsius    

    assert (temp_max.shape == \
            temp_min.shape == \
            temp.shape == \
            rh.shape)
    
    print(f'\nCalculating VPD [kPa] from Temperature and RH...\n')

    # doi: 10.1175/BAMS-86-2-225
    temp_dew = temp - ((100 - rh) / 5) 

    # ISBN: 978-92-5-104219-9 
    es_tmin = 0.6108 * np.exp((17.27 * temp_min) / (temp_min + 237.3))
    es_tmax = 0.6108 * np.exp((17.27 * temp_max) / (temp_max + 237.3))

    es = (es_tmax + es_tmin) / 2

    ea = 0.6108 * np.exp((17.27 * temp_dew) / (temp_dew + 237.3))

    # kPa
    vpd = es - ea

    if isinstance(vpd, xr.DataArray): vpd.rename('VPD')

    return vpd



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

def axgrid(ax: plt.Axes, 
           color: str = 'dimgray', 
           ls: str = '--', 
           which: str = 'major', 
           visible: bool = True, 
           alpha: float = 0.4):

    ax.grid(which = which, 
            color = color, 
            visible = visible, 
            ls = ls, 
            alpha = alpha, 
            zorder = 0)


def common_free_lims(ax: plt.Axes, 
                     xs: np.ndarray | pd.Series | pd.DataFrame, 
                     ys: np.ndarray | pd.Series | pd.DataFrame, 
                     xerr: np.ndarray | pd.Series | pd.DataFrame, 
                     yerr: np.ndarray | pd.Series | pd.DataFrame, 
                     rel_b: float = 0.1):

    if isinstance(ys, pd.Series): ys = ys.values
    if isinstance(xs, pd.Series): xs = xs.values
    if isinstance(yerr, pd.Series): yerr = yerr.values
    if isinstance(xerr, pd.Series): xerr = xerr.values

    imax = np.nanmax(np.concatenate([xs + xerr, 
                                     ys + yerr]))
    imin = np.nanmin(np.concatenate([xs - xerr, 
                                     ys - yerr]))

    max_rel = np.nanmax(np.abs([imax, imin]))

    imax_rel = imax + max_rel * rel_b
    imin_rel = imin - max_rel * rel_b

    ax.set_ylim((imin_rel, imax_rel))
    ax.set_xlim((imin_rel, imax_rel))

    return imin_rel, imax_rel


def bar_lims(ax, xs, ys, rel_b: float = 0.1):

    import numpy as np
    import pandas as pd

    if isinstance(ys, pd.Series): ys = ys.values
    if isinstance(xs, pd.Series): xs = xs.values

    xmax                    = np.max(xs)
    xmin                    = np.min(xs)

    ymax                    = np.max(ys)
    ymin                    = np.min(ys)

    xabs                    = np.max(np.abs([xmax, xmin]))
    yabs                    = np.max(np.abs([ymax, ymin]))

    xmax_rel                = xmax + xabs * rel_b + 0.5
    xmin_rel                = xmin - xabs * rel_b - 0.5

    ymax_rel                = ymax + yabs * rel_b
    ymin_rel                = 0 - yabs * rel_b

    ax.set_ylim((ymin_rel, ymax_rel))
    ax.set_xlim((xmin_rel, xmax_rel))

    return [xmin_rel, xmax_rel], [ymin_rel, ymax_rel]


def free_lims(ax: plt.Axes, 
              xs: np.ndarray | pd.Series | pd.DataFrame, 
              ys: np.ndarray | pd.Series | pd.DataFrame, 
              rel_b: float = 0.1):

    if isinstance(ys, pd.Series): ys = ys.values
    if isinstance(ys, pd.DataFrame): ys = ys.values
    if isinstance(xs, pd.Series): xs = xs.values
    if isinstance(xs, pd.DataFrame): xs = xs.values

    xmax                    = np.nanmax(xs)
    xmin                    = np.nanmin(xs)

    ymax                    = np.nanmax(ys)
    ymin                    = np.nanmin(ys)

    xabs                    = np.nanmax(np.abs([xmax, xmin]))
    yabs                    = np.nanmax(np.abs([ymax, ymin]))

    xmax_rel                = xmax + xabs * rel_b + 0.5
    xmin_rel                = xmin - xabs * rel_b - 0.5

    ymax_rel                = ymax + yabs * rel_b
    ymin_rel                = ymin - yabs * rel_b

    ax.set_ylim((ymin_rel, ymax_rel))
    ax.set_xlim((xmin_rel, xmax_rel))

    return [xmin_rel, xmax_rel], [ymin_rel, ymax_rel]
    

def free_date_lim(ax, ts, x = True, y = False, rel_b: float = 0.1):

    import pandas as pd
    import numpy as np
    import datetime as dt

    ts.dropna()

    tsmax                   = np.max(ts)
    tsmin                   = np.min(ts)

    if x: ax.set_xlim(tsmin, tsmax)
    if y: ax.set_ylim(tsmin, tsmax)

    return tsmin, tsmax


def free_numeric_lim(ax, xs, x = False, y = True, rel_b: float = 0.1):

    import pandas as pd
    import numpy as np

    if isinstance(xs, pd.Series): xs = xs.values
    if isinstance(xs, pd.DataFrame): xs = xs.values

    tsmax                   = np.nanmax(xs) 
    tsmin                   = np.nanmin(xs)

    tsabs                   = np.max(np.abs([tsmax, tsmin]))

    tsmax_rel               = tsmax + tsabs * rel_b 
    tsmin_rel               = tsmin - tsabs * rel_b

    if x: ax.set_xlim((tsmin_rel, tsmax_rel))
    if y: ax.set_ylim((tsmin_rel, tsmax_rel))

    return tsmin_rel, tsmax_rel

    

def numeric_ticks(ax, 
                  nticks_y = 5, 
                  nticks_x = 5, 
                  fs_ticks = 10, 
                  integerx = False, 
                  integery = False,
                  time_x: bool = False, 
                  x: bool = True, 
                  y: bool = True):

    from matplotlib.ticker import MaxNLocator

    xlocator = MaxNLocator(prune = 'both', 
                           nbins = nticks_x, 
                           integer = integerx)
    ylocator = MaxNLocator(prune = 'both', 
                           nbins = nticks_y, 
                           integer = integery)

    ax.yaxis.set_major_locator(ylocator)
    ax.xaxis.set_major_locator(xlocator)

    if not x: ax.set_xticklabels([])
    if not y: ax.set_yticklabels([])

    ax.tick_params(axis = 'both', which = 'major', labelsize = fs_ticks)
    


def numeric_ticks_2(ax, nticks_y = 5, nticks_x = 5, fs_ticks = 10, integerx = False, integery = False,
                  time_x: bool = False, x: bool = True, y: bool = True):

    from matplotlib.ticker import MaxNLocator
    import matplotlib.dates as mdates

    if time_x:
        
        xlocator = mdates.AutoDateLocator(minticks = nticks_x)
        ax.xaxis.set_major_formatter(mdates.AutoDateFormatter(xlocator)) 


    else:
        
        xlocator            = MaxNLocator(prune = 'both', nbins = nticks_x, integer = integerx)

    ylocator                = MaxNLocator(prune = 'both', nbins = nticks_y, integer = integery)

    ax.yaxis.set_major_locator(ylocator)
    ax.xaxis.set_major_locator(xlocator)



    if not x: ax.set_xticklabels([])
    if not y: ax.set_yticklabels([])

    ax.tick_params(axis = 'both', which = 'major', labelsize = fs_ticks)
    




import pandas as pd
import re

def index(y0: int, 
          y1: int, 
          t_res: str,
          month0: int = 1,
          month1: int = 12, 
          day0: int = 1,
          day1: int = 31,
          leapday: bool = True,
          range_kwargs: dict = {},
          resample_kwargs: dict = {}) -> pd.Series:

    """
    Create time index from beginning of year 0
    to end of year 0;
    To use for pandas time series;
    """
    
    if t_res[-1] == 'H': end_time = str(24 - int(t_res[:-1]))
    else: end_time = '23:59:00'

    y0_time = f'{y0}-{month0:02}-{day0:02} 00:00:00'
    y1_time = f'{y1}-{month1:02}-{day1:02} {end_time}'

    t_res_comp = re.split('(\d+)', t_res)

    if t_res_comp[0].isdigit(): t_res_num = t_res_comp[0]
    else: t_res_num = None

    t_res_str = t_res_comp[-1]

    time_raw = pd.date_range(y0_time, 
                             y1_time, 
                             freq = t_res_str,
                             **range_kwargs).to_series()

    # Make yearly groups for resample
    time_groups = time_raw.groupby(time_raw.index.year, 
                                   group_keys=False)

    # Resample. Not always needed (but e.g. for multiples 
    # '8D' frequecies)
    time_resampled = time_groups.resample(t_res, 
                                          **resample_kwargs) \
                                          .asfreq()
    
    if leapday == False: 
        time_resampled = drop_leapday(time_resampled)

    return time_resampled


def drop_leapday(series):

    """
    Drop the leapday entries from any
    time series. I think there should be
    an easier way as long as index is 
    datetime.
    """

    mask                        = [False if (s.month == 2 and s.day == 29) else True for s in series.index]

    series_out                  = series[mask]

    return series_out


def FLX_to_datetime(frequencies: list = []):

    translate                   = { 'Y': 'Y',
                                    'M': 'M',
                                    'D': 'D',
                                    'H': 'H'
                                   }
    
    freqs_translated            = [translate[freq] for freq in frequencies]
    
    return freqs_translated


def lowest_frequency(frequencies: list = []):

    """
    pandas date time offset strings

    """

    import numpy as np

    scores                      = {'Y': 0, 'M': 1, 'D': 2, 'H': 3}
    scores_rev                  = {v: k for k, v in scores.items()}

    freq_scores                 = [scores[freq] for freq in frequencies]

    freq_scores_array           = np.array(freq_scores)

    lowest_freq_score           = np.min(freq_scores_array)

    lowest_freq                 = scores_rev[lowest_freq_score]

    return lowest_freq


def index_to_datetime(df: pd.DataFrame, 
                      format: str | None = None):

    import pandas as pd

    df.index                    = pd.to_datetime(df.index, format = format)

    return df




import matplotlib.pyplot as plt
import numpy as np
from cartopy import crs
import pandas as pd

def init_xy(ax: plt.Axes, 
            xs: np.ndarray | pd.Series | pd.DataFrame, 
            ys: np.ndarray | pd.Series | pd.DataFrame, 
            xerr: np.ndarray | pd.Series | pd.DataFrame | None = None, 
            yerr: np.ndarray | pd.Series | pd.DataFrame | None = None, 
            axhv_color: str = 'k', 
            axhv_ls: str = '--', 
            axhv_lw: float = 0.7, 
            axhv_alpha: float = 0.8, 
            axhv_dashes: tuple = (4, 4),
            diag_color: str = 'k', 
            diag_ls: str = '--', 
            diag_lw: float = 0.7, 
            diag_alpha: float = 0.8, 
            diag_dashes: tuple = (4,4),
            title: str = '', 
            xlabel: str = '', 
            ylabel: str = '', 
            y_title: float = 1.1, 
            fs_title: int = 14, 
            fs_label: int = 12, 
            fs_ticks: int = 10,
            ax_tag: str | None = None, 
            ax_tag_x: float = 0.5, 
            ax_tag_y: float = 1.0):
    
    from my_.plot.limits import axgrid, common_free_lims, numeric_ticks
    from my_.plot.style import nospines

    axgrid(ax)
    nospines(ax)

    plt.suptitle(title, fontsize = fs_title, y = y_title)
    
    ax.set_ylabel(ylabel, fontsize = fs_label)
    ax.set_xlabel(xlabel, fontsize = fs_label)
    
    ax.tick_params(axis='both', which='major', pad=10)

    ax.axhline(0, color = axhv_color, 
                ls = axhv_ls, 
                lw = axhv_lw, 
                alpha = axhv_alpha, 
                dashes = axhv_dashes, 
                zorder = 1)
    
    ax.axvline(0, color = axhv_color, 
                ls = axhv_ls, 
                lw = axhv_lw, 
                alpha = axhv_alpha, 
                dashes = axhv_dashes, 
                zorder = 1)

    ax.plot([-1.1, 1.1], [-1.1, 1.1], 
            transform = ax.transAxes, 
            color = diag_color, 
            ls = diag_ls, 
            lw = diag_lw, 
            alpha = diag_alpha, 
            dashes = diag_dashes, 
            zorder = 1)
    
    props = dict(facecolor='white', edgecolor='none', alpha=0.85)

    if ax_tag is not None:

        ax.text(ax_tag_x, 
                ax_tag_y, 
                ax_tag, 
                fontsize = fs_label, 
                transform = ax.transAxes, 
                va = 'bottom', 
                ha = 'center', 
                bbox = props)
    
    ax.axis('square')
    
    common_free_lims(ax = ax, 
                     xs = xs, 
                     ys = ys,
                     xerr = xerr,
                     yerr = yerr)

    numeric_ticks(ax, fs_ticks = fs_ticks)


def init_bar(ax, xs, ys, title: str = '', fs_title: float = 14, y_title: float = 1.0, ylabel: str = '', xlabel: str = '', fs_label: float = 12,
             axhv_color: str = 'k', axhv_ls: str = '--', axhv_lw: float = 0.7, axhv_alpha: float = 0.8, axhv_dashes: tuple = (4, 4),
             ax_tag = '', ax_tag_x: float = 0.5, ax_tag_y: float = 1.0, fs_ticks: float = 10,):

    from my_.plot.limits import axgrid, bar_lims, numeric_ticks
    from my_.plot.style import nospines

    import matplotlib.pyplot as plt

    axgrid(ax)
    nospines(ax)

    plt.suptitle(title, fontsize = fs_title, y = y_title)
    
    ax.set_ylabel(ylabel, fontsize = fs_label)
    ax.set_xlabel(xlabel, fontsize = fs_label)

    ax.axhline(0, color = axhv_color, 
                ls = axhv_ls, lw = axhv_lw, 
                alpha = axhv_alpha, dashes = axhv_dashes, zorder = 0)
    
    props = dict(facecolor='white', edgecolor='none', alpha=0.85)

    ax.text(ax_tag_x, ax_tag_y, ax_tag, fontsize = fs_label, transform = ax.transAxes, va = 'bottom', ha = 'center', bbox = props)

    bar_lims(ax, xs, ys)

    numeric_ticks(ax, nticks_x = len(ys), nticks_y = 5, fs_ticks = fs_ticks, integerx = True, x = False)


def EU3_plot_init(rotnpole_lat: float = 39.25, 
                  rotnpole_lon: float = -162.0, 
                  semmj_axis: int = 6370000, 
                  semmn_axis: int = 6370000, 
                  lon_extents: list = [351.1, 57], 
                  lat_extents: list = [27, 65.7]):

    rp = crs.RotatedPole(pole_longitude = rotnpole_lon,
                         pole_latitude = rotnpole_lat,
                         globe = crs.Globe(semimajor_axis = semmj_axis,
                         semiminor_axis = semmn_axis))
    
    pc = crs.PlateCarree()

    xs, ys, _ = rp.transform_points(pc, np.array(lon_extents), np.array(lat_extents)).T
    
    return rp, pc, xs, ys


def init_ts(ax: plt.Axes, 
            xs: np.ndarray | pd.Series | pd.DataFrame, 
            ys: np.ndarray | pd.Series | pd.DataFrame, 
            axhv_color: str = 'k', 
            axhv_ls: str = '--', 
            axhv_lw: float = 0.7, 
            axhv_alpha: float = 0.8, 
            axhv_dashes: tuple = (4, 4),
            title: str = '', 
            xlabel: str = '', 
            ylabel: str = '', 
            y_title: float = 1.1, 
            fs_title: int = 14, 
            fs_label: int = 12, 
            fs_ticks = 10,
            ax_tag = '', 
            ax_tag_x: float = 0.5, 
            ax_tag_y: float = 1.0):
    
    from my_.plot.style import nospines
    from my_.plot.limits import axgrid, numeric_ticks, free_lims

    axgrid(ax)
    nospines(ax)

    plt.suptitle(title, 
                 fontsize = fs_title, 
                 y = y_title)
    
    ax.set_ylabel(ylabel, 
                  fontsize = fs_label)
    
    ax.set_xlabel(xlabel, 
                  fontsize = fs_label)
    
    ax.tick_params(axis='both', 
                   which='major', 
                   pad=10)

    ax.axhline(0, color = axhv_color, 
               ls = axhv_ls, 
               lw = axhv_lw, 
               alpha = axhv_alpha, 
               dashes = axhv_dashes, 
               zorder = 1)
    
    ax.axvline(0, color = axhv_color, 
                ls = axhv_ls, 
                lw = axhv_lw, 
                alpha = axhv_alpha, 
                dashes = axhv_dashes, 
                zorder = 1)
    
    props = dict(facecolor = 'white', 
                 edgecolor = 'none', 
                 alpha = 0.85)

    ax.text(ax_tag_x, 
            ax_tag_y, 
            ax_tag, 
            fontsize = fs_label, 
            transform = ax.transAxes, 
            va = 'bottom', 
            ha = 'center', 
            bbox = props)

    numeric_ticks(ax, fs_ticks = fs_ticks)
    free_lims(ax, xs, ys)


def init_lol(ax: plt.Axes, 
             xs: np.ndarray | pd.Series | pd.DataFrame, 
             ys: np.ndarray | pd.Series | pd.DataFrame, 
             axhv_color: str = 'k', 
             axhv_ls: str = '--', 
             axhv_lw: float = 0.7, 
             axhv_alpha: float = 0.8, 
             axhv_dashes: tuple = (4, 4),
             title: str = '', 
             xlabel: str = '', 
             ylabel: str = '', 
             y_title: float = 1.1, 
             fs_title: int = 14, 
             fs_label: int = 12, 
             fs_ticks = 10,
             ax_tag = '', 
             ax_tag_x: float = 0.5, 
             ax_tag_y: float = 1.0):
    
    from my_.plot.style import nospines
    from my_.plot.limits import axgrid, numeric_ticks, free_lims, bar_lims

    axgrid(ax)
    nospines(ax)

    ax.set_title(title, 
                 fontsize = fs_title, 
                 y = y_title)
    
    ax.set_ylabel(ylabel, 
                  fontsize = fs_label)
    
    ax.set_xlabel(xlabel, 
                  fontsize = fs_label)
    
    ax.tick_params(axis='both', 
                   which='major', 
                   pad=10)
    
    props = dict(facecolor = 'white', 
                 edgecolor = 'none', 
                 alpha = 0.8)

    ax.text(ax_tag_x, 
            ax_tag_y, 
            ax_tag, 
            fontsize = fs_label, 
            transform = ax.transAxes, 
            va = 'bottom', 
            ha = 'center', 
            bbox = props)

    numeric_ticks(ax, 
                  nticks_x = 4, 
                  nticks_y = len(ys) + 1, 
                  fs_ticks = fs_ticks, 
                  integery = True,
                  integerx = True)
    

def init_ts_2(ax, xs, ys, axhv_color: str = 'k', axhv_ls: str = '--', axhv_lw: float = 0.7, axhv_alpha: float = 0.8, axhv_dashes: tuple = (4, 4),
       title: str = '', xlabel: str = '', ylabel: str = '', y_title: float = 1.1, fs_title: int = 14, fs_label: int = 12, fs_ticks = 10,
       ax_tag = '', ax_tag_x: float = 0.5, ax_tag_y: float = 1.0, grid: bool = True, spines: bool = False):
    
    from my_.plot.style import nospines
    from my_.plot.limits import axgrid, numeric_ticks_2, free_date_lim, free_numeric_lim

    import matplotlib.pyplot as plt

    if grid: axgrid(ax)
    if not spines: nospines(ax)

    plt.suptitle(title, fontsize = fs_title, y = y_title)
    
    ax.set_ylabel(ylabel, fontsize = fs_label)
    ax.set_xlabel(xlabel, fontsize = fs_label)
    
    ax.tick_params(axis='both', which='major', pad=10)

    ax.axhline(0, color = axhv_color, 
                ls = axhv_ls, lw = axhv_lw, 
                alpha = axhv_alpha, dashes = axhv_dashes, zorder = 0)
    
    ax.axvline(0, color = axhv_color, 
                ls = axhv_ls, lw = axhv_lw, 
                alpha = axhv_alpha, dashes = axhv_dashes, zorder = 0)
    
    props = dict(facecolor='white', edgecolor='none', alpha=0.85)

    ax.text(ax_tag_x, ax_tag_y, ax_tag, fontsize = fs_label, transform = ax.transAxes, va = 'bottom', ha = 'center', bbox = props)

    numeric_ticks_2(ax, fs_ticks = fs_ticks, time_x = True)
    free_date_lim(ax, xs)
    free_numeric_lim(ax, ys)



def init_dist(ax: plt.axes,
                xs: np.ndarray | pd.Series | pd.DataFrame,
                ys: np.ndarray | pd.Series | pd.DataFrame,
                axhv_color: str = 'k',
                axhv_ls: str = '--', 
                axhv_lw: float = 0.7, 
                axhv_alpha: float = 0.8, 
                axhv_dashes: tuple = (4, 4),
                title: str = '', 
                xlabel: str = '', 
                ylabel: str = '', 
                y_title: float = 1.1, 
                fs_title: float = 14.0, 
                fs_label: float = 12.0, 
                fs_ticks: float = 10.0,
                ax_tag: str = '', 
                ax_tag_x: float = 0.5, 
                ax_tag_y: float = 1.0) -> None:
    
    from my_.plot.style import nospines
    from my_.plot.limits import axgrid, numeric_ticks, free_lims

    axgrid(ax)
    nospines(ax)

    plt.suptitle(title, fontsize = fs_title, y = y_title)
    
    ax.set_ylabel(ylabel, fontsize = fs_label)
    ax.set_xlabel(xlabel, fontsize = fs_label)
    
    ax.tick_params(axis='both', which='major', pad=10)

    ax.axhline(0, color = axhv_color, 
                ls = axhv_ls, lw = axhv_lw, 
                alpha = axhv_alpha, dashes = axhv_dashes, zorder = 0)
    
    ax.axvline(0, color = axhv_color, 
                ls = axhv_ls, lw = axhv_lw, 
                alpha = axhv_alpha, dashes = axhv_dashes, zorder = 0)
    
    props = dict(facecolor='white', edgecolor='none', alpha=0.85)

    ax.text(ax_tag_x, ax_tag_y, ax_tag, fontsize = fs_label, transform = ax.transAxes, va = 'bottom', ha = 'center', bbox = props)

    numeric_ticks(ax, fs_ticks = fs_ticks)
    free_lims(ax, xs, ys)


def init_annotation_ax(axs, x : float = 0.05, y : float = 1.05, fs = 12):

    import string
    
    abc                     = string.ascii_lowercase
    
    list_abc                = list(abc)

    for iax, ax in enumerate(axs):

        ax.text(x, y, list_abc[iax] + ')', fontsize = fs,
                transform = ax.transAxes,va = 'bottom', ha = 'center')


def init_pie(ax, fs_label = 10, ax_tag = '', ax_tag_x: float = 0.5, ax_tag_y: float = 0.95):
    
    props = dict(facecolor='white', edgecolor='none', alpha=0.85)

    ax.text(ax_tag_x, ax_tag_y, ax_tag, fontsize = fs_label, transform = ax.transAxes, va = 'bottom', ha = 'center', bbox = props)


import pandas as pd
import numpy as np
import pymannkendall as mk

def mann_kendall(data: np.ndarray | pd.Series,
                 type: str = 'original_test',
                 **kwargs):

    # https://pypi.org/project/pymannkendall/
    
    test = getattr(mk, type)

    results = test(data, **kwargs)

    return results




import os
from my_.files.handy import create_dirs, check_file_exists
from glob import glob
import xarray as xr
import numpy as np


all_EUCORDEX_daily = {
    'name': 'BGC_EU3',
    'version': (1, 0, 6),
    'path': '/p/scratch/cjibg31/jibg3105/data/ERA5L/all/',
    'type_file': 'netcdf',
    'year_start': 1979,
    'month_start': 1,
    'year_end': 2022,
    'month_end': 12,
    'resolution_time': 'D',
    'grid': 'ERA5L_EUCORDEX',
    'variables': ['P',
                  'ET',
                  'PET',
                  'Runoff',
                  'SM'],
    'variable_names': {'P': 'tp',
                       'ET': 'e',
                       'PET': 'pev',
                       'Runoff': 'ro',
                       'SM': ['swvl1', 'swvl2', 'swvl3', 'swvl4']},
    'variable_dimensions': {'P': ['time', 'lat', 'lon'], 
                            'ET': ['time', 'lat', 'lon'],
                            'PET': ['time', 'lat', 'lon'],
                            'Runoff': ['time', 'lat', 'lon'],
                            'SM': ['time', 'layer', 'lat', 'lon']}, 
    'variable_units': {'P': 'm/day',
                       'ET': 'm/day',
                       'PET': 'm/day',
                       'Runoff': 'm/day',
                       'SM': 'm^3/m^3'},
    'mask_value': -32767,
}


def create_yearly_files(path_rawdata: os.PathLike,
                        path_out: os.PathLike):
    
    """
    Input directory should only contain the ERA5L rawdata files...
    """

    print('\nConvert ERA5L files to yearly...\n')
    
    create_dirs(path_out)

    files = sorted(glob(f'{path_rawdata}/*.nc'))

    for f in files:

        file_year = f.split('/')[-1]

        if check_file_exists(f'{path_out}/{file_year}'): continue

        data_raw = xr.open_dataset(f)

        data_raw['e'] = data_raw['e'] * -1
        data_raw['pev'] = data_raw['pev'] * -1

        print(f'Create yearly file for file {f}...\n')

        data_raw.to_netcdf(f'{path_out}/{file_year}',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])
        

def merge_files(paths_rawdata: os.PathLike,
                path_out: os.PathLike):
    
    """
    Input directory should only contain the ERA5L rawdata files...
    """

    print('\nMerge ERA5L files...\n')
    
    create_dirs(path_out)
    
    files = []

    for p in paths_rawdata:
    
        pp = sorted(glob(f'{p}/*.nc'))

        files.extend(pp)

    data_raw = xr.open_mfdataset(files)

    years = np.unique(data_raw.time.dt.year.values)

    for y in years:

        print(f'Create yearly file for year {y}...\n')

        if check_file_exists(f'{path_out}/{y}.nc'): continue

        data_y = data_raw.sel(time = data_raw.time.dt.year.isin([y]))

        data_y.to_netcdf(f'{path_out}/{y}.nc',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])




def src_var_insitu_df(name_case: str, sources = [], variables = [],
                        file_format: str = 'csv',
                        path_stations: str = 'user_in/csv/',
                        file_stations: str = 'stations.csv',
                        year_start: int = 1995, year_end: int = 2018):
    
    from glob import glob
    import pandas as pd

    from my_.resources.sources import query_variables
    from my_.files.handy import create_dirs

    for src in sources:

        dir_out                 = f'out/{name_case}/{file_format}/'
        
        create_dirs([dir_out])
        
        file_out                = f'{dir_out}/Insitu_{src}.{file_format}'

        if glob(file_out): print(f'Output file for {name_case} - {src} is already available.\n'); continue
        
        print(f'Output file for {src} is not available and will be created.\n')

        df_stations             = pd.read_csv(f'{path_stations}/{file_stations}')

        processing              = query_variables(src, 'processing')

        if processing == 'ONEFLUX':

            oneflux(name_case, df_stations, src, variables, qc_values = {'M': 0.8, 'D': 0.8, 'H': 1.0},
                    file_out = file_out, file_format = file_format,
                    year_start = year_start, year_end = year_end)
        
        if processing == 'ICOS-ETC-L2':
            
            oneflux(name_case, df_stations, src, variables, qc_values = {'M': 0.8, 'D': 0.8, 'H': 1.0},
                    file_out = file_out, file_format = file_format,
                    year_start = year_start, year_end = year_end,
                    dir_prefix = 'ICOSETC', file_prefix = 'ICOSETC_')
            
        if processing == 'ARCHIVE':
            
            archive(name_case, df_stations, src, variables,
                    file_out = file_out, file_format = file_format,
                    year_start = year_start, year_end = year_end,
                    dir_prefix = 'ICOSETC', file_prefix = 'ICOSETC_')

        
    
def oneflux(name, df_stations, src: str, variables = [], resample_method = 'mean', 
            qc_flag: bool = True, qc_values: dict = {'M': 0.8, 'D': 0.8, 'H': 1.0},
            file_format: str = 'csv', file_out: str = 'oneflux_out.csv',
            year_start: int = 1995, year_end: int = 2018,
            dir_prefix = 'FLX', file_prefix = 'FLX_*_FULLSET_'):

    from my_.misc.list_dict_key import keys_same_value, translate_dict_values, filter_dict_keys
    from my_.resources.sources import query_variables, available_variables
    from my_.files.csv import open_csv 
    from my_.files.handy import save_df
    from my_.series.time import index, lowest_frequency, index_to_datetime
    from my_.series.group import  multi_columns
    from my_.series.interpolate import resample, reindex
    from my_.series.aggregate import mask_ge, mask_le, concat
    from my_.resources.units import transform_df

    from pathlib import Path

    path                        = query_variables(src, 'path')
    var_names                   = query_variables(src, 'var_names')
    time_steps                  = query_variables(src, 'var_time_step')
    qcs                         = query_variables(src, 'var_QC')
    csv_open_args               = query_variables(src, 'csv_open_args')
    date_formats                = query_variables(src, 'date_format')

    vars_avail, _               = available_variables(src, variables)

    names_stations              = df_stations['name'].values
    names_FLX_stations          = df_stations['FLX_name'].values
    landcover_stations          = df_stations['landcover'].values

    filtered_ts                 = filter_dict_keys(time_steps, vars_avail)
    grouped_ts                  = keys_same_value(filtered_ts)

    grouped_ts_src              = translate_dict_values(idict = grouped_ts, idict_translate = var_names)
    grouped_ts_qc               = translate_dict_values(idict = grouped_ts, idict_translate = qcs)

    list_ts                     = list(grouped_ts_src)
    lowest_freq_ts              = lowest_frequency(list_ts)

    time                        = index(y0 = year_start, y1 = year_end, t_res = lowest_freq_ts)

    list_dfs                    = []
   
    for i_station in range(len(names_stations)):

        name                    = names_stations[i_station]
        lc                      = landcover_stations[i_station]
        name_FLX                = names_FLX_stations[i_station]

        print(f'Extract site {name}, in-situ ONEFLUX data for {src}...\n')

        for ts in list_ts:
        
            file                = f'{path}/{dir_prefix}*{name_FLX}*/{file_prefix}*{ts}{ts}*.csv'
        
            ts_df               = open_csv(file, **csv_open_args)

            ts_df_date          = index_to_datetime(ts_df, format = date_formats[ts])
        
            df_sel              = ts_df_date[grouped_ts_src[ts]]
            
            if qc_flag:

                qc_value        = qc_values[ts]
                    
                df_sel_qc       = ts_df_date[grouped_ts_qc[ts]]

                if ts == 'H':

                    df_sel      = mask_le(df_sel, df_sel_qc, qc_value)
                
                if ((ts == 'D') or (ts == 'M')):

                    df_sel      = mask_ge(df_sel, df_sel_qc, qc_value)

            df_sel_resampled    = resample(df_sel, lowest_freq_ts, resample_method)

            df_sel_reindexed    = reindex(df_sel_resampled, time)


            df_sel_transformed  = transform_df(df_sel_reindexed, src, grouped_ts[ts])
            
            new_cols            = multi_columns([[src], grouped_ts[ts], [lc], [name]], 
                                                names = ['Source', 'Variable', 'Landcover', 'Station'])

            df_sel_transformed.columns = new_cols

            list_dfs.append(df_sel_transformed)

    df_oneflux                  = concat(list_dfs, sort = False)

    save_df(df_oneflux, file_out, file_format)


def archive(name, df_stations, src: str, variables = [], 
            file_format: str = 'csv', file_out: str = 'oneflux_out.csv',
            year_start: int = 2017, year_end: int = 2024,
            dir_prefix = 'FLX', file_prefix = 'FLX_*_FULLSET_'):
    
    import pandas as pd

    from my_.misc.list_dict_key import keys_same_value, translate_dict_values, filter_dict_keys
    from my_.resources.sources import query_variables, available_variables
    from my_.files.csv import open_csv 
    from my_.files.handy import save_df
    from my_.series.time import index, index_to_datetime
    from my_.series.group import  multi_columns
    from my_.series.interpolate import resample, reindex
    from my_.series.aggregate import concat
    from my_.resources.units import transform_df
    from my_.files.handy import check_file_exists

    path = query_variables(src, 'path')    
    csv_open_args = query_variables(src, 'csv_open_args')
    date_format = query_variables(src, 'date_format')

    names_stations = df_stations['name'].values
    names_FLX_stations = df_stations['FLX_name'].values
    landcover_stations = df_stations['landcover'].values

    cols = pd.MultiIndex.from_tuples([], names=('Source', 'Variable', 'Landcover', 'Station'))

    time_index = index(y0 = year_start,
                       y1 = year_end,
                       t_res = 'D')

    df_oneflux = pd.DataFrame(columns = cols,
                              index = time_index)
   
    for i_station in range(len(names_stations)):

        name = names_stations[i_station]
        lc = landcover_stations[i_station]
        name_FLX = names_FLX_stations[i_station]

        print(f'Extract site {name}, in-situ ONEFLUX data for {src}...\n')

        file = f'{path}/{dir_prefix}*{name_FLX}*ARCHIVE_L2/{file_prefix}*ANCILLARY_L2.csv'

        if not check_file_exists(file): continue
        
        ts_df = open_csv(file, **csv_open_args)

        for v in variables:

            ts_df_values = ts_df[ts_df['VARIABLE'] == v]['DATAVALUE']
                                 
            ts_df_dates = ts_df[ts_df['VARIABLE'].isin([f'{v}_DATE', f'{v}_DATE_START'])]['DATAVALUE']

            ts_df_dates = pd.to_datetime(ts_df_dates, format = date_format)

            new_cols = multi_columns([[src], [v], [lc], [name]], 
                                      names = ['Source', 'Variable', 'Landcover', 'Station'])
                                
            df_i = pd.DataFrame(columns = new_cols, 
                                index = ts_df_dates.values, 
                                dtype = 'float')

            df_i[(src, v, lc, name)] = ts_df_values.values

            df_i_transformed = transform_df(df_i, src, variables)

            df_oneflux = pd.merge(df_oneflux, 
                                  df_i_transformed,
                                  left_index = True,
                                  right_index = True,
                                  how = 'outer')

    df_i_t_agg = df_oneflux.groupby(level = 0).mean()

    save_df(df_i_t_agg, file_out, file_format)


if __name__ == '__main__':

    from argparse import ArgumentParser

    parser                      = ArgumentParser()

    parser.add_argument('--name', '-n', help = 'Name for your work', type = str)
    parser.add_argument('--sources', '-s', nargs = '+', help = 'Names of source data', type = str)
    parser.add_argument('--variables', '-v', nargs = '+', help = 'List of variables', type = str)
    parser.add_argument('--file_format', '-ff', help = 'file format suffix for dataframe', type = str, default = 'csv')
    parser.add_argument('--path_stations', '-ps', help='Path of the station information file', type = str, default = 'user_in/csv/')
    parser.add_argument('--file_stations', '-fs', help='File name of the station information file', type = str, default = 'stations.csv')
    parser.add_argument('--year_start', '-y0', help='Start year', type = int, default = 1995)
    parser.add_argument('--year_end', '-y1', help='End year', type = int, default = 2018)

    args                        = parser.parse_args()

    src_var_insitu_df(name_case = args.name, sources = args.sources, variables = args.variables,
                        file_format = args.file_format, path_stations = args.path_stations,
                        file_stations = args.file_stations, year_start = args.year_start, year_end = args.year_end)


from my_.files.handy import create_dirs, check_file_exists
import os
from glob import glob
import xarray as xr
import numpy as np

v381_yearly = {
    'name': 'GLEAM_v3.8a_yearly',
    'version': (3, 8, 1),
    'path': '/p/scratch/cjibg31/jibg3105/data/GLEAM/3.8a/yearly/',
    'type_file': 'netcdf',
    'year_start': 1980,
    'month_start': 1,
    'year_end': 2022,
    'month_end': 12,
    'resolution_time': 'YE',
    'grid': 'GLEAM',
    'variables': ['ET', 
                  'PET',
                  'SM',],
    'variable_names': {'ET': 'E',
                       'PET': 'Ep',
                       'SM': ['SMsurf', 'SMroot'],
                       },
    'variable_dimensions': {'ET': ['time', 'lat', 'lon'],
                            'PET': ['time', 'lat', 'lon'],
                            'SM': ['time', 'layer', 'lat', 'lon'],},
    'variable_units': {'ET': 'mm/year',
                       'PET': 'mm/year',
                       'SM': 'm^3/m^3'},
    'mask_value': None
}


def create_yearly_files(path_rawdata: os.PathLike,
                        path_out: os.PathLike):
    
    """
    Input directory should only contain the GLEAM rawdata files...
    """

    print('\nConvert GLEAM files to yearly...\n')
    
    create_dirs(path_out)

    files = glob(f'{path_rawdata}/*GLEAM*.nc')

    data_raw = xr.open_mfdataset(files)

    years = np.unique(data_raw.time.dt.year.values)

    for y in years:

        print(f'Create yearly file for year {y}...\n')

        if check_file_exists(f'{path_out}/{y}.nc'): continue

        data_y = data_raw.sel(time = data_raw.time.dt.year.isin([y]))

        data_y.to_netcdf(f'{path_out}/{y}.nc',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])



import pandas as pd
import numpy as np

def gauss_kde_pdf(data: pd.DataFrame | pd.Series | np.ndarray , 
                    n: int = 1000, 
                    return_dict: bool = False):
    
    import pandas as pd
    import numpy as np
    from scipy.stats import gaussian_kde

    if isinstance(data, pd.DataFrame): data = data.values
    if isinstance(data, pd.Series): data = data.values

    data_clean = data[~np.isnan(data)]

    mins = np.min(data_clean)

    maxs = np.max(data_clean)

    xs = np.linspace(mins, maxs, n)

    kde_sp = gaussian_kde(data_clean)

    ys = kde_sp.pdf(xs)

    if return_dict:     
        return pd.DataFrame({'xs': xs, 'ys': ys})
    else:
        return xs, ys


def gauss_kde_cdf(data, n = 1000):

    import pandas as pd
    import numpy as np
    from scipy.stats import gaussian_kde
    from scipy.special import ndtr

    if isinstance(data, pd.DataFrame): data = data.values
    if isinstance(data, pd.Series): data = data.values


    data_clean = data[~np.isnan(data)]

    mins = np.min(data_clean)

    maxs = np.max(data_clean)

    xs = np.linspace(mins, maxs, n)

    kde_sp = gaussian_kde(data_clean)

    ys = np.array([ndtr(np.ravel(x - kde_sp.dataset) / kde_sp.factor).mean() for x in xs])

    return xs, ys


def distribution_pdf(distribution: str = 'norm', parameter: int | list = [0, 1], n = 100000):

    import scipy.stats as stats
    import numpy as np

    func_dist = getattr(stats, distribution)

    mins = func_dist.ppf(1/n, *parameter)

    maxs = func_dist.ppf(1 - 1/n, *parameter)

    xs = np.linspace(mins, maxs, n)

    ys = func_dist.pdf(xs, *parameter)

    return xs, ys


def distribution_cdf(distribution: str = 'norm', 
                     parameter: int | list = [0, 1], 
                     n = 100000):

    import scipy.stats as stats
    import numpy as np

    func_dist = getattr(stats, distribution)


    mins = func_dist.ppf(1/n, *parameter)

    maxs = func_dist.ppf(1 - 1/n, *parameter)

    xs = np.linspace(mins, maxs, n)

    ys = func_dist.cdf(xs, *parameter)

    return xs, ys

def distribution_fit(array, distribution: str = 'gamma'):

    import scipy.stats as stats

    if distribution == 'gaussian_kde': return array

    func_dist                       = getattr(stats, distribution)

    parameters                      = func_dist.fit(array)

    return parameters


def distribution_data(distribution: str = 'norm', 
                      parameter: dict | tuple = {}, 
                      n = 100000):

    #Baustelle

    import scipy.stats as stats
    import numpy as np

    func_dist = getattr(stats, distribution)

    data = func_dist.rvs(0, 1, n)    




def open_parquet(file_str, args = {}):

    import pandas as pd
    from glob import glob

    file                        = glob(file_str)[0]

    df                          = pd.read_parquet(file, **args)

    return df